Require Import Arith.
Require Import NPeano.
Require Import Omega. 
Require Import List.
Require Classes.EquivDec.
Require Import LIFO.List.Misc.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Map.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Mem.
Require Import LIFO.List.Assoc.
Require Import LIFO.Option.Option.
Require CertifiedBSML.Primitives.Specification.
Require CertifiedBSML.Primitives.Properties.
Require CertifiedBSML.Support.ParSig.
Require Import Program.

Set Implicit Arguments.

Module Type TYPE
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (PropertiesM : CertifiedBSML.Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(PropertiesM)).

  Import BsmlPrimitives PropertiesM ParSigM.

  Obligation Tactic := program_simpl.

  Definition replicate (A:Type)(a:A) :
    { vr : par A | forall (i:processor), get vr i = a } :=
    mkpar(fun _ => a).

  Definition lreplicate(A:Type)(a:A) : 
    forall (i:processor), get (proj1_sig (replicate a)) i = a
    := proj2_sig (replicate a).

  Hint Rewrite lreplicate: bsmlbase.

  Obligation Tactic := program_simpl.

  Program Definition parfun(A B: Type)(f:A->B)(v:par A) :
    { vr: par B | forall (i:processor), get vr i = f (get v i) }:=
    apply (replicate f) v.
  Next Obligation.
    autorewrite with bsml; reflexivity.
  Defined.

  Definition lparfun(A B: Type)(f:A->B)(v:par A) := 
    proj2_sig (parfun f v).

  Hint Rewrite lparfun : bsmlbase.

  Ltac rewrite_parfun :=
    let g := fresh "f" in 
      let H := fresh "H" in 
    match goal with 
      | [ |- ?E ] =>
        match E with 
          | context t[proj1_sig(parfun ?f ?v)] => 
            set(g:=f)
        end
    end;
    set(H:= lparfun g);
      unfold g in *;
        rewrite H;
          clear H g.

  Ltac bsmlbase := 
    repeat (rewrite_parfun; autorewrite with bsmlbase par_sig option).

  Program Definition parfun2(A B C: Type)(f:A->B->C)(v1:par A)(v2: par B) :
    { vr: par C | forall (i:processor), get vr i = f (get v1 i) (get v2 i) }:=
    apply (parfun f v1) v2.
  Next Obligation.
    autorewrite with bsml; reflexivity.
  Defined.

  Definition lparfun2(A B C: Type)(f:A->B->C)(v1:par A)(v2:par B) := 
    proj2_sig (parfun2 f v1 v2).

  Hint Rewrite lparfun2 : bsmlbase.

  Program Definition 
    mask (A:Type)(p:processor -> bool)(v1 v2: par A) :
    { vr: par A | forall i:processor, 
      get vr i = if (p i) then get v1 i else get v2 i } :=
    apply(apply(mkpar(fun i x y =>if(p i)then x else y))v1)v2.
  Next Obligation.
    intros; autorewrite with bsml; reflexivity.
  Qed.

  Definition lmask (A B:Type)(prop:processor->bool)(v1 v2: par A):=
    proj2_sig(mask prop v1 v2).

  Program Definition 
    parfunif (A B:Type)(prop:processor->bool)(f1 f2:A->B)(v: par A):
    { vr : par B | forall i:processor, 
      get vr i = if (prop i) then f1(get v i) else f2(get v i) } :=
    apply(mkpar(fun i=>if (prop i) then f1 else f2)) v.
  Next Obligation.
    autorewrite with bsml; destruct(prop i); reflexivity.
  Qed.

  Definition lparfunif (A B:Type)(prop:processor->bool)(f1 f2:A->B)(v: par A):=
    proj2_sig(parfunif prop f1 f2 v).

  Program Definition 
    applyif (A B:Type)(prop:processor->bool)(f1 f2:par(A->B))(v: par A):
    { vr : par B | forall i:processor, 
      get vr i = if (prop i) then (get f1 i)(get v i) else (get f2 i)(get v i) } :=
    apply(apply(apply(mkpar(fun i f1 f2 v=>if (prop i) then f1 v else f2 v))f1)f2)v.
  Next Obligation.
    autorewrite with bsml; destruct(prop i); reflexivity.
  Qed.

  Definition lapplyif (A B:Type)(prop:processor->bool)(f1 f2:par(A->B))(v: par A):=
    proj2_sig(applyif prop f1 f2 v).

  Program Definition 
    parfunat(A B:Type)(n:processor)(f1 f2:A->B)(v: par A) :
    { vr : par B | forall i:processor, get vr i = 
      if beq_nat n i then f1 (get v i) else f2 (get v i) } :=
    parfunif (beq_nat n) f1 f2 v.

  Definition lparfunat(A B:Type)(n:processor)(f1 f2:A->B)(v:par A) := 
    proj2_sig(parfunat n f1 f2 v).

  Program Definition 
    applyat(A B:Type)(n:processor)(f1 f2:par(A->B))(v: par A) :
    { vr : par B | forall i:processor, get vr i = 
      if beq_nat n i then (get f1 i)(get v i) else (get f2 i)(get v i) } :=
    applyif (beq_nat n) f1 f2 v.

  Definition lapplyat(A B:Type)(n:processor)(f1 f2:par(A->B))(v:par A) := 
    proj2_sig(applyat n f1 f2 v).

  Hint Rewrite lmask lapplyif lapplyat lparfunif lparfunat : bsmlbase.

  Definition this: {X : par processor | forall i : processor, get X i =  i} := mkpar id.
  

  Definition procs := PropertiesM.processors.

  Program Definition 
    get_one_aux (A:Type)(data: par A)(src: par nat ) :
    { vr: par (option A) | forall i:processor, 
      get vr i = Some (get data (processorOfNatMod (get src i))) } := 
    let src_processors := parfun processorOfNatMod src in 
      let ask := put(parfun
        (fun pid dst=> if beq_nat (proj1_sig pid) (proj1_sig dst) then Some tt else None)
        src_processors) in
        let replace_by_data := 
          parfun2(fun f data dst => match (f dst) with 
                                      | Some _ => Some data 
                                      | None => None 
                                    end) in
          let reply := put(replace_by_data ask data) in 
            apply reply src_processors.
  Next Obligation.
    autorewrite with bsml;
      repeat (rewrite <- beq_nat_refl);
        auto.
  Defined.

  Definition lget_one_aux(A:Type)(data:par A)(src:par nat) := 
    proj2_sig (get_one_aux data src). 

  Hint Rewrite lget_one_aux : bsmlbase.

  Obligation Tactic := idtac.

  Program Definition 
    get_one (A:Type)(data: par A)(src: par nat ) :
    { vr: par A | forall i:processor, 
      get vr i = get data (processorOfNatMod(get src i)) } := 
    parfun (@noSome A) (parSig (get_one_aux data src) (fun x=>x<>None) _).
  Next Obligation.
    intros A data src i;
      autorewrite with bsmlbase par_sig option;
        discriminate.
  Defined.
  Next Obligation.
    intros A data src i;
      apply someInjectiveEq;
        bsmlbase; trivial.
  Defined.

  Definition lget_one(A:Type)(data:par A)(src:par nat) := 
    proj2_sig (get_one data src). 

  Hint Rewrite lget_one: bsmlbase.

  Program Definition 
    get_list_aux 
    (A:Type)(data: par A)(src: par (list nat)):
    { vr: par (list (option A)) | forall i:processor, 
      get vr i = List.map (fun i =>Some(get data (processorOfNatMod i))) (get src i) } := 
    let src_processors := parfun (List.map processorOfNatMod) src in 
      let ask := put(parfun
        (fun l dst=> if mem Classes.EquivDec.nat_eq_eqdec (proj1_sig dst) (listProj l) 
          then Some tt else None)
        src_processors) in
        let replace_by_data := 
          parfun2(fun f data dst => match (f dst) with 
                                      | Some _ => Some data 
                                      | None => None 
                                    end) in
          let reply := put(replace_by_data ask data) in 
            parfun2 (fun f l=>List.map f l) reply src_processors.
  Next Obligation.
    program_simpl; autorewrite with bsml.
      rewrite map_map. apply map_assumption;
        intros a Ha; autorewrite with bsml;
          unfold listProj; simpl; repeat (rewrite map_map).
    unfold processorOfNatMod; simpl;
      apply in_map with (f:=fun x=> x mod bsp_p) in Ha.
        assert(mem Classes.EquivDec.nat_eq_eqdec (a mod bsp_p) 
          (map(fun x:nat=>x mod bsp_p) (get src i)) = true) 
          as Hmem by (apply inMemEq; assumption);
            rewrite Hmem; auto.
  Defined.

  Definition lget_list_aux (A:Type)(data: par A)(src: par (list nat)):=
    proj2_sig(get_list_aux data src).

  Hint Rewrite lget_list_aux: bsmlbase.

  Program Definition 
    get_list 
    (A:Type)(data: par A)(src: par (list nat)):
    { vr: par (list A) | forall i:processor, 
      get vr i = List.map (fun i => (get data (processorOfNatMod i))) (get src i) } := 
    parfun (List.map (@noSome A))(parInSig (get_list_aux data src) _ _).
  Next Obligation.
    intros A data src i x;
      autorewrite with bsmlbase par_sig option; intro H;
        rewrite in_map_iff in H; destruct H as [a Ha]; destruct Ha as [Ha1 Ha2];
          rewrite <- Ha1;
            discriminate.
  Defined.
  Next Obligation.
    intros A data src i;
      autorewrite with bsmlbase par_sig option;
        apply mapInjectiveEq with (f:=Some); try apply someInjectiveEq;
          repeat rewrite map_map;
            rewrite map_ext with (g:=@proj1_sig (option A) (fun x=>x<>None)).
              set(H:=inSigProj); unfold listProj in H. rewrite H.
                autorewrite with bsmlbase par_sig option.
                  auto.
              intro a; destruct a; apply someNoSomeId.
  Defined.

  Definition lget_list (A:Type)(data: par A)(src: par (list nat)):=
    proj2_sig(get_list data src).

  Hint Rewrite lget_list: bsmlbase.

  Program Definition 
    put_one (A:Type)(data: par (nat*A)) : 
    { vr : par (list A) | forall i, forall v, In v (get vr i) <-> exists j, get data j = (i,v) } :=
    let msg := parfun 
      (fun (dstpidValue:(nat*A)) dst =>
        if beq_nat (fst dstpidValue) (proj1_sig dst) 
          then Some (snd dstpidValue) else None)
      data in
    parfun (fun f => mapFilterSome f processors) (put msg).
  Next Obligation.
    intros A data msg i v; split.
    (* -> *)
    bsmlbase;
      rewrite inMapSomeEq; rewrite mapFilterSomeProp; rewrite filter_In; rewrite in_map_iff;
        intro H; destruct H as [H _]; destruct H as [src H];
          unfold msg in H; generalize dependent H; autorewrite with bsml; bsmlbase; intro H;
            exists src.
    destruct H; case_eq (beq_nat (fst (get data src)) (`i)); intro Hi; rewrite Hi in H.
      injection H; clear H; intro H; 
        rewrite beq_nat_true_iff in Hi;
          destruct (get data src); simpl in *; subst; reflexivity.
      discriminate.
    (* <- *)
    bsmlbase;
    rewrite inMapSomeEq; rewrite mapFilterSomeProp; rewrite filter_In; rewrite in_map_iff;
      intro H; destruct H as [j Hj];
        split; auto; exists j; split.
          unfold msg; autorewrite with bsml; bsmlbase;
            destruct (get data j); inversion Hj; subst; simpl; rewrite <- beq_nat_refl; split; auto.
          apply inProcessors.
  Defined.

  Definition lput_one (A:Type)(data:par(nat*A)) := proj2_sig (put_one data).

  Hint Rewrite put_one : bsmlbase.

  Program Definition 
    put_list (A:Type)(data: par(list(nat*A))) : 
    { vr : par (list A) | forall i, forall v, 
      In v (get vr i) <-> exists j, inFirst (proj1_sig i,v) (get data j) } :=
    let msg := parfun 
      (fun listDstValue dst => assoc Classes.EquivDec.nat_eq_eqdec (proj1_sig dst) listDstValue)
      data in
      parfun (fun f => mapFilterSome f processors) (put msg).
  Next Obligation.
    intros A data msg i v; split;
      bsmlbase; rewrite inMapSomeEq; rewrite mapFilterSomeProp; rewrite filter_In; rewrite in_map_iff.
    (* -> *)
    intro H; destruct H as [H _]; destruct H as [src H];
      unfold msg in H; generalize dependent H; autorewrite with bsml; bsmlbase; intro H;
        exists src; destruct H; rewrite <- assocInFirst; eassumption.
    (* <- *)
    intro H; destruct H as [j Hj]; split; auto; exists j; split.
      unfold msg; autorewrite with bsml; bsmlbase; rewrite assocInFirst; assumption.
      apply inProcessors.
  Defined.

  Definition lput_list (A:Type)(data:par(list(nat*A))) := proj2_sig (put_list data).

  Hint Rewrite put_list : bsmlbase.

End TYPE.

Module Make
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (PropertiesM : CertifiedBSML.Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(PropertiesM)) :
  TYPE(BsmlPrimitives)(PropertiesM)(ParSigM).

  Include TYPE(BsmlPrimitives)(PropertiesM)(ParSigM).

End Make.