Require Import Program.
Require Export Coq.Classes.Init.
Require Export Coq.Classes.RelationClasses.

Generalizable All Variables.

(** *Type classes for abstract algebra concepts *)

(** Unit  *)
Class Neutral_old `{equ : Equivalence A eqA} (op: A -> A -> A) (e : A) :=
{
  left_neutral_old  : forall a, eqA (op e a) a;
  right_neutral_old : forall a, eqA (op a e)  a
}.

(** Unit  *)
Class NeutralLeft `{equ : Equivalence A eqA} (op: A -> A -> A) (e : A) :=
{
  left_neutral  : forall a, eqA (op e a) a
}.

Class NeutralRight `{equ : Equivalence A eqA} (op: A -> A -> A) (e : A) :=
{
  right_neutral : forall a, eqA (op a e)  a
}.

Class Neutral `{equ : Equivalence A eqA} (op: A -> A -> A) (e : A) :=
{
    NEUTRAL_left_neutral  :>  NeutralLeft op e;
    NEUTRAL_right_neutral :>  NeutralRight op e
}.

(** Commutativity  *)
Class  Commutative `{equ : Equivalence A eqA} (op: A -> A -> A) :={
  commutative : forall (x y: A), eqA (op  x y) (op y x)
}.

(** Associativity  *)
Class  Associative `{equ : Equivalence A eqA} (op: A -> A -> A) :={
  associative : forall (x y z: A), eqA (op (op x y) z) (op x (op y z))
}.

(** Semi-Group *)
Class SemiGroup  `{equ : Equivalence A eqA} (op : A->A->A) :=
{
  SemiGroup_Associative :> Associative op
}.

(** Quasi-Group *)
Class QuasiGroup `{equ : Equivalence A eqA} (op : A->A->A) :=
{
  quasiGroupLaw : forall a b :A, exists  !x, exists !y, eqA (op a x)  b /\ eqA (op y a) b
}.

(** Monoid definition  *)
Class Monoid   `{equ : Equivalence A eqA} (op : A->A->A) (e : A)  := 
{
  Monoid_Associative :> Associative op; 
  Monoid_Neutral :> Neutral op e
}.

(** Commutative Monoid definition  *)
Class CommutativeMonoid `{equ : Equivalence A eqA} (op : A->A->A) (e : A)  := 
{
  CommutativeMonoid_Commutative :> Commutative op;
  CommutativeMonoid_Monoid :>  Monoid op e
}.

(** Group definition  *)
Class Group  `{equ : Equivalence A eqA} (op : A->A->A)   := 
{ groupNeutral : A;
  Group_Associative :> Associative op; 
  Group_Neutral :> Neutral op groupNeutral ;
  groupInverse : forall x:A, exists x', eqA (op x x') groupNeutral
}.

(** Commutative Group definition  *)
Class CommutativeGroup  `{equ : Equivalence A eqA} (op : A->A->A) (e : A)  := 
{
  CommutativeGroup_Commutative :> Commutative op; 
  CommutativeGroup_Group :> Group op 
}.

(** Multiplicative Group definition  *)
Class MultiplicativeGroup  `{equ : Equivalence A eqA} (op : A->A->A)   := 
{ multiplicativeGroupNeutral : A; 
  nilA : A;
  MultiplicativeGroup_Associative :> Associative op; 
  MultiplicativeGroup_Neutral :> Neutral op multiplicativeGroupNeutral ;
  multiplicativeGroupInverse : forall x:A, exists x', ~ eqA x nilA -> eqA (op x x') multiplicativeGroupNeutral
}.

(**  Distributivity of otime over oplus  *)
Class Distributivity  `{equ : Equivalence A eqA} `(otimes : A->A->A) (oplus  : A->A->A):={
  leftDistributivity : forall (x y z : A), 
    eqA (otimes x (oplus y z))  (oplus (otimes x y) (otimes x z));
  rightDistributivity : forall (x y z : A), 
    eqA (otimes (oplus y z) x)  (oplus (otimes y x) (otimes z x))
}.

(** SemiRing definition *)
Class SemiRing `{equ : Equivalence A eqA} `(oplus : A->A->A) (otimes : A->A->A)  (ep et : A) :=
{
  SemiRing_Monoid_p :> CommutativeMonoid oplus ep;
  SemiRing_Monoid_t :> Monoid otimes et;
  SemiRing_Distributivity :> Distributivity otimes oplus;
  semiRingNil : forall a, eqA (otimes ep a) ep /\ eqA (otimes a ep) ep 
}.

(** Ring definition *)
Class Ring  `{equ : Equivalence A eqA} (oplus : A->A->A) (otimes : A->A->A)  (ep et : A) :=
{
  Ring_CommutativeGroup_p :> CommutativeGroup oplus ep;
  Ring_Monoid_t :> Monoid otimes et;
  Ring_Distributivity :> Distributivity otimes oplus
}.

(** Commutative Ring definition *)
Class CommutativeRing  `{equ : Equivalence A eqA} (oplus : A->A->A) (otimes : A->A->A)  (ep et : A) :=
{
  CommutativeRing_CommutativeGroup_p :> CommutativeGroup oplus ep;
  CommutativeRing_Monoid_t :> Monoid otimes et;
  CommutativeRing_Commutative_t :> Commutative otimes;
  CommutativeRing_Distributivity :> Distributivity otimes oplus
}.

(** Commutative Ring definition *)
Class Field  `{equ : Equivalence A eqA} (oplus : A->A->A) (otimes : A->A->A)  (ep et : A) :=
{
  Field_CommutativeGroup_p :> CommutativeGroup oplus ep;
  Field_Group_t :> MultiplicativeGroup otimes ;
  Fiel_Distributivity :> Distributivity otimes oplus
}.

Program Instance commutativeMonoid_monoid  `(CommutativeMonoid)  : 
Monoid op e.

Program Instance group_is_monoid  `(Group)  : Monoid op groupNeutral.

Program Instance commutativeRing_is_Ring  `(CommutativeRing)  :  Ring  oplus otimes ep et.

(* Section semiRingNat. *)

Require Import Arith.

(**  0 is neutral for addition in nat *)
Program Instance plus_neutral : Neutral plus 0.
Solve Obligations  using (constructor;auto).

(**  addition in nat is commutative *)
Program Instance commutative_plus :  Commutative plus.
Solve Obligations  using (auto with arith).

(**  addition in nat is associative *)
Program Instance associative_plus :  Associative plus.
Solve Obligations  using (auto with arith).

(**  nat with addition is a commutative monoid *)
Program Instance commutativeMonoid_plus : CommutativeMonoid plus 0.
Next Obligation.
  constructor;typeclasses eauto.
Qed.

(**  multiplication in nat is associative *)
Program Instance associative_mult : Associative mult.
Solve Obligations  using (auto with arith).

(**  1 is neutral for multiplication in nat *)
Program Instance neutral_mult : Neutral mult 1.
Solve Obligations  using (constructor;auto with arith).

(**  multiplication in nat is associative *)
Program Instance commutative_mult : Commutative mult.
Solve Obligations  using (auto with arith).

(**  nat with multiplication is a Commutative Monoid *)
Program Instance commutativeMonoid_mult : CommutativeMonoid mult 1.
Next Obligation.
  constructor; typeclasses eauto.
Qed.


(**  multiplication is distributive over addition in nat *)
Program Instance distributive_mult_plus : Distributivity mult plus.
Solve Obligations  using (auto with arith).
Next Obligation.
       repeat rewrite (commutative (op:=mult) x).
       auto with arith.
Qed.

(**  nat with multiplication and addition is a SemiRing *)
Program Instance SemiRing_nat : SemiRing plus mult 0 1.

(* End semiRingNat. *)



(** *Z is a ring*)
(* Section anneauZ. *)

Require Import ZArith.
Open Scope Z_scope.

(**  0 is neutral for addition in Z *)
Program Instance Zplus_neutral : Neutral (eqA:=eq) Zplus 0.
Solve Obligations  using (constructor;auto with zarith).

(**  addition in Z is commutative *)
Program Instance Zplus_commutative :  Commutative (eqA:=eq) Zplus.
Solve Obligations  using (intros;auto with zarith).

(**  addition in Z is associative *)
Program Instance Zplus_associative :  Associative (eqA:=eq) Zplus.
Solve Obligations  using (intros;auto with zarith).

(**  Z with addition is a group *)
Program Instance group_Zplus : Group (eqA:=eq) Zplus.
Solve Obligations  using (typeclasses eauto || intro x;exists (-x); auto with zarith ).

(**  Z with addition is a commutative group *)
Program Instance commutativeGroup_Zplus : CommutativeGroup (eqA:=eq) Zplus 0.
(* Solve Obligations  using (typeclasses eauto). (*  *) *)

(**  multiplication in Z is associative *)
Program Instance associative_zmult : Associative (eqA:=eq) Zmult.
Solve Obligations  using (auto with zarith).

(**  1 is neutral for multiplication in Z *)
Program Instance neutral_zmult : Neutral (eqA:=eq) Zmult 1.
Solve Obligations  using (constructor; auto with zarith).


(**  Z with multiplication is a Monoid *)
Program Instance monoid_zmult : Monoid (eqA:=eq) Zmult 1.
(* Solve Obligations  using (typeclasses eauto). *)

(**  multiplication is distributive over addition in Z *)
Program Instance distributive_zmult_zplus : Distributivity (eqA:=eq) Zmult Zplus.
Solve Obligations  using (auto with zarith).

(**  Z with multiplication and addition is a Ring *)
Program Instance Ring_Z : Ring (eqA:=eq) Zplus Zmult 0 1.
(* Solve Obligations  using (typeclasses eauto). *)
Close Scope Z_scope.
(* End anneauZ. *)

(* Section semiAnneauBool. *)
Require Import Bool.
Open Scope bool_scope.

Program Instance or_neutral : Neutral orb false. 
Solve Obligations  using (constructor; auto with bool).

Program Instance or_commutative :  Commutative orb.
Solve Obligations  using (auto with bool).

Program Instance or_associative :  Associative  orb.
Solve Obligations  using (auto with bool).

Program Instance commutative_monoid_or : CommutativeMonoid orb false.

Solve Obligations  using (constructor; typeclasses eauto).
(* Solve Obligations  using (typeclasses eauto). *)

Program Instance and_neutral : Neutral andb true.
Solve Obligations  using (constructor; auto).
Next Obligation.
  constructor;
  unfold andb; intros; case a; reflexivity.
Qed. 

Program Instance and_commutative :  Commutative andb.
Solve Obligations  using (auto with bool).

Program Instance and_associative :  Associative  andb.
Solve Obligations  using (auto with bool).

Program Instance commutative_monoid_and : CommutativeMonoid andb true.
Solve Obligations  using (constructor; typeclasses eauto).


Program Instance distributive_andb_orb : Distributivity andb orb.
Solve Obligations  using (intros x y z; case x; case y ;  case z ; unfold orb; reflexivity).

Program Instance distributive_orb_andb : Distributivity orb andb.
Solve Obligations  using (intros x y z; case x; case y ;  case z ; auto with bool).

Program Instance SemiRing_bool : SemiRing andb orb true false.
Solve Obligations  using (typeclasses eauto || intro a; case a; split; auto).

Close Scope bool_scope.
(* End semiAnneauBool. *)


(* Section FieldQ. *)
Require Import QArith. 
Open Scope Q_scope.

Program Instance Q_neutral : Neutral (eqA:=Qeq) Qplus 0.
Next Obligation.
  constructor; apply Qplus_0_l.
Qed.
Next Obligation.
  constructor; apply Qplus_0_r.
Qed.

Program Instance Qplus_commutative :  Commutative  (eqA:=Qeq) Qplus.
Next Obligation.
  apply Qplus_comm.
Qed.

Program Instance Qplus_associative :  Associative  Qplus.
Solve Obligations  using (intros;symmetry;apply Qplus_assoc).

Program Instance group_Qplus : Group Qplus. 
Next Obligation.
exists (-x). apply (Qplus_opp_r x).
Qed.

Program Instance commutativeGroup_Qplus : CommutativeGroup Qplus 0.
(* Solve Obligations  using (typeclasses eauto || idtac). *)

Program Instance Qmult_neutral : Neutral Qmult 1.
Solve Obligations  using (constructor ; (apply Qmult_1_r||apply Qmult_1_l)).

Program Instance Qmult_commutative :  Commutative Qmult.
Solve Obligations  using (apply Qmult_comm).

Program Instance Qmult_associative :  Associative  Qmult.
Solve Obligations  using (symmetry; apply Qmult_assoc).

Program Instance multiplicativeGroup_Qmult : MultiplicativeGroup Qmult :={nilA:=0}.
Solve Obligations  using (typeclasses eauto|| idtac).
Next Obligation.
  exists (/x).
  apply Qmult_inv_r.
Qed.

(* Program Instance commutative_group_Qmult : CommutativeGroup Qmult 1. *)
(* Solve Obligations  using (typeclasses eauto|| idtac). *)
(* Next Obligation. *)


Program Instance distributive_Qmult_Qplus : Distributivity Qmult Qplus.
Next Obligation.
  apply Qmult_plus_distr_r.
Qed.
Next Obligation.
  apply Qmult_plus_distr_l.
Qed.


Program Instance Field_bool : Field Qplus Qmult 0 1.
(* Solve Obligations  using (typeclasses eauto || intro a; case a; split; auto). *)
Close Scope Q_scope.

(* End FieldQ. *)

(* (* MAX *) *)

(* Program Instance max_neutral : Neutral max  0. *)
(* Solve  Obligations using  (constructor;kintro b; apply max_comm; unfold max; simpl; reflexivity). *)

(* Program Instance max_commutative :  Commutative  max. *)
(* Solve Obligations  using (intros; apply max_comm). *)

(* Program Instance max_associative :  Associative  max. *)
(* Solve Obligations  using (intros; symmetry; apply max_assoc). *)


(* Program Instance plus_neutral : Neutral plus  0. *)
(* (* Solve Obligations  using (constructor;auto with arith). *) *)

(* Program Instance plus_commutative :  Commutative plus. *)
(* Solve Obligations  using (intros;auto with arith). *)

(* Program Instance plus_associative :  Associative  plus. *)
(* Solve Obligations  using (intros;auto with arith). *)

(* Program Instance distrib_max_plus : Distributivity  plus max. *)
(* Next Obligation. *)
(*    induction x;  induction y; induction z; simpl in *; (reflexivity || auto with arith). *)
(* Qed. *)
  


(* Program Instance  commutativeMonoid_max_0: CommutativeMonoid max 0. *)
(* Solve Obligations using (typeclasses eauto). *)

(* Program Instance  commutativeMonoid_sum_0: CommutativeMonoid plus 0. *)
(* Solve Obligations using (typeclasses eauto). *)

(* Program Instance semiRing_max_plus: SemiRing max plus 0 0. *)
(* Solve Obligations using (typeclasses eauto ||idtac). *)
(* Next Obligation.  *)
(* split. *)