Require Import List.
Require Import LIFO.List.Notations.
Require Import LIFO.List.App.
Require Import LIFO.Algebra.AlgebraClasses.
(** list destruction principle for homomorphism
list is either nil, or a singleton or a concatenation of two non-empty lists *)  
Lemma listDestruct : forall A (l : list A),
  {l=nil}+{exists a, l=a::nil}+{exists b, exists a,exists la,exists lb,l=(a::la)++(b::lb)}.
  Proof.
    intros A0 l.
    induction l.
    left;left. reflexivity. 
    destruct IHl.
      destruct s. 
        subst; left ; right; apply ex_intro with (x:=a); reflexivity.
        right. elim e. intros x lsingleton. rewrite lsingleton. apply ex_intro with (x:=x). apply ex_intro with (x:=a).
          apply ex_intro with (x:=nil : list A0). apply ex_intro with (x:=nil : list A0). simpl. reflexivity.
        right. 
          elim e;clear e;intros b0 e; elim e;clear e;intros a0 e; elim e;clear e;intros la e; elim e;clear e;intros lb e.
          apply ex_intro with (x:=b0); apply ex_intro with (x:=a); apply ex_intro with (x:=(a0::la)); 
            apply ex_intro with (x:=lb).
          subst. simpl. reflexivity.
  Qed.
  Implicit Arguments listDestruct [A].
  
Tactic Notation "destructList" constr(l) "as" ident(H) :=
    destruct (listDestruct l) as [H | H];[destruct H as [H | H];[|destruct H as [a H]] |destruct H as [b H]; destruct H as [a H] ; destruct H as [la H] ;destruct H as [lb H] ] .

Lemma homomorphism_induction : forall A, forall P : list A -> Prop, 
  P nil 
  -> (forall a, P (a::nil))
    -> (forall l l', P l ->P l' -> P (l++l'))
      -> forall l,  P l.
Proof.
  intros A P H H0 H1 l.
  induction l.
  apply H.
  replace (a::l) with ((a::nil) ++l) by reflexivity.
  apply H1.
  apply H0.
  apply IHl.
Qed.



Generalizable All Variables.

(** * Homomorphism Classes  *)

(** Homomorphism classes with all elements of the homomorphism as parameter  *)
Class Homomorphism_F {A B: Type} `{eqb : Equivalence B eqB}  `(h : list A -> B)  `(op: B -> B -> B) `(f:A -> B)  :=
{
  Homomorphism_F_Associativity :> Associative  op ;
  Homomorphism_F_nil_id :>  Neutral op (h nil);
  Homomorphism_F_list_destruct : forall (x y: list A), eqB (h (x++y)) (op (h x) (h y));
  Homomorphism_F_singleton : forall (a: A), eqB (h [a]) (f a)
}.


(** Homomorphism classes with only the function  as parameter, decomposition is done in the fields  *)
Class Homomorphism {A B: Type} `{equB : Equivalence B eqB}  (h : list A -> B)   :=
{
  Homo_op: B -> B -> B;
  Homo_f :A -> B;
  Homomorphism_Associativity :> Associative  Homo_op ;
  Homomorphism_nil_id :>  Neutral Homo_op (h nil);
  Homomorphism_list_destruct : forall (x y: list A), eqB (h (x++y)) (Homo_op (h x) (h y));
  Homomorphism_singleton : forall (a: A), eqB (h [a]) (Homo_f a)
}.

(** Homomorphism classes with  the function and the operator as parameter,  
   function on singleton is build using h on the singleton list  *)
Class Homomorphism' {A B: Type} `{eqb : Equivalence B eqB}  `(op: B -> B -> B) `(h:list A -> B)  :=
{
  Homomorphism'_Associativity :> Associative  op ;
  Homomorphism'_nil_id :>  Neutral op (h nil);
  Homomorphism'_list_destruct : forall (x y: list A), eqB (h (x++y)) (op (h x) (h y));
  Homomorphism'_f := (fun a => h (a::nil)) : A -> B
}.


(** Homomorphism classes with  the function and the operator and the necessary operator properties as parameter,
   function on singleton is build using h on the singleton list  *)
Class Homomorphism'' {A B: Type} `{eqb : Equivalence B eqB} (e:B) (op: B -> B -> B)  {opass:Associative  op}   {opneutral :Neutral op e} (h:list A -> B)   :=
{
  Homomorphism''_h_nil : h nil = e;
  Homomorphism''_list_destruct : forall (x y: list A), eqB (h (x++y)) (op (h x) (h y));
  Homomorphism''_f := (fun a => h (a::nil)) : A -> B
}.

(* /!\ Warning /!\  
   adding an instance linking  Homomorphism''  and  Homomorphism' leads to infinite loops in type class resolution   
*)


Global Program Instance  homoF_homo `( Homomorphism_F )  : Homomorphism h:={Homo_op := op; Homo_f:=f}.
Next Obligation.
  destruct H. auto.
Qed.
Next Obligation.
  destruct H. cbv. auto.
Qed.




Global Program Instance homo'_homo `(Homomorphism' A B eqb op h) : Homomorphism h.
Next Obligation.
  exact(h [X]).
Defined.
Next Obligation.
destruct H. auto.
Qed.
Next Obligation.
 destruct H; cbv. reflexivity.
Qed.


Lemma fold_homomorphism   `(Homomorphism A B eq  h) : forall l l' a, h (l++[a]++l') = (Homo_op (h:=h)) (fold_left (Homo_op (h:=h)) (map (Homo_f (h:=h)) l) (h[]) ) (Homo_op (h:=h) (Homo_f (h:=h) a) (fold_left (Homo_op (h:=h)) (map (Homo_f (h:=h)) l') (h[]) )).
Proof.
  induction l ; induction l'.
  intro a; simpl. cbv. destruct H. rewrite left_neutral. rewrite right_neutral. auto.
  intro a0; simpl. rewrite left_neutral. replace  (a0::a::l') with ([a0]++[]++[a]++l') by (simpl; reflexivity).
             rewrite Homomorphism_list_destruct.  rewrite IHl'. simpl. rewrite left_neutral.
             rewrite Homomorphism_singleton. 
             Require Import LIFO.List.Fold.
             rewrite left_neutral. Unset Printing All . 
             replace (Homo_f (h:=h) a) with (Homo_op (h:=h) (Homo_f (h:=h) a) ( h [])) by (apply right_neutral).             
             destruct H.
             rewrite (fold_assoc_prop (f:=Homo_op)). rewrite right_neutral.
             reflexivity.

intros a0. simpl.  replace  (a::l++[a0]) with ([a]++l++[a0]++[]) by (simpl; reflexivity).
              rewrite Homomorphism_list_destruct. rewrite IHl.
              repeat rewrite left_neutral.  repeat rewrite right_neutral. 
              replace (Homo_f (h:=h) a) with (Homo_op (h:=h) (Homo_f (h:=h) a) ( h [])) by (apply right_neutral).            
              rewrite (fold_assoc_prop (f:=Homo_op)). 
              rewrite associative.
              rewrite Homomorphism_singleton.
              reflexivity.

intros a1. simpl .  replace  (a::l++a1::a0::l') with ([a]++l++[a1]++a0::l') by (simpl; reflexivity).
               rewrite Homomorphism_list_destruct. rewrite IHl.
              simpl; repeat rewrite left_neutral.  
              replace (Homo_f (h:=h) a) with (Homo_op (h:=h) (Homo_f (h:=h) a) ( h [])) by (apply right_neutral).            
              rewrite (fold_assoc_prop (f:=Homo_op)). 
              rewrite associative.
              rewrite Homomorphism_singleton.
              reflexivity.
Qed.

Lemma fold_homomorphism_2 `(Homomorphism A B eq  h) : 
  forall l l' a, 
    fold_left (Homo_op (h:=h)) (l++l') a   =  Homo_op (h:=h) (fold_left (Homo_op (h:=h)) l a) ( fold_left (Homo_op (h:=h)) l' (h[])).
Proof.
induction  l; intros a'. (* as [ | a' l'].  *)
   
     simpl.  intros a.
     rewrite <- (fold_assoc_prop (f:=Homo_op)).  rewrite right_neutral. reflexivity.
     
     simpl. intros; rewrite IHl. reflexivity.
Qed.



Lemma fold_homomorphism_3 `(Homomorphism A B eq  h) : 
  forall l, h l = fold_left (Homo_op (h:=h)) (map (Homo_f (h:=h)) l) (h[]).
Proof.
 induction l.
      reflexivity.
      simpl.  rewrite left_neutral. 
         replace (Homo_f (h:=h) a)  with (Homo_op (h:=h) (Homo_f (h:=h) a) (h[]) ) by (rewrite right_neutral; reflexivity). 
         rewrite (fold_assoc_prop (f:=Homo_op)).         
         replace (a::l) with ([a]++l) by reflexivity.         
         rewrite Homomorphism_list_destruct.
         rewrite Homomorphism_singleton. 
         rewrite  IHl. 
         reflexivity.
Qed.


(**  Identity is an homomorphism  *)
Global Program Instance id_is_an_homomorphism (A:Type): Homomorphism'(eqB:=eq) (app (A:=A)) id.
Next Obligation.
  unfold id.
  typeclasses eauto.
Qed.

(** Fold left can be an Homomorphism if the operator is associative  *)
Global Program  Instance foldleft_is_an_homomorphism''
  {A:Type}
  (op: A->A->A)
  (e: A)
  {Ha:Associative  op}
  {Hu: Neutral  op e} : 
  Homomorphism'' e op (fun l =>List.fold_left op l e)|300 .
Next Obligation.
  induction x.
  simpl. rewrite left_neutral.  reflexivity.
  simpl. rewrite left_neutral.
  assert(forall a y,   fold_left op y a = op a (fold_left op y e)).
    intros a0 y0.
    generalize dependent a0.
    induction y0.
    simpl. intros a0. rewrite right_neutral. reflexivity.
    intros a1.    
    simpl. rewrite IHy0.  rewrite  associative.
        rewrite <- IHy0. rewrite left_neutral; reflexivity.
        rewrite H. rewrite (H a x). rewrite associative. rewrite IHx.
  reflexivity.
Qed.

Global Program  Instance foldleft_is_an_homomorphism 
  {A:Type}
  (op: A->A->A)
  (idop: A)
  (Ha:Associative  op) 
  (Hu: Neutral  op idop) : 
  Homomorphism' op (fun l =>List.fold_left op l idop).
Next Obligation.
  destruct Hu as [HuL HuR].
  destruct HuL as [HuL].
  destruct HuR as [HuR].
  induction x.
  simpl. rewrite HuL. reflexivity.
  simpl. rewrite HuL. 
  destruct Ha. 
  assert(forall a y,   fold_left op y a = op a (fold_left op y idop)).
    intros a0 y0.
    generalize dependent a0.
    induction y0. auto.
    simpl. rewrite HuL. rewrite IHy0. intros. rewrite <- associative.
    rewrite <- IHy0. reflexivity.
  rewrite H. rewrite (H a x). rewrite associative. rewrite IHx.
  reflexivity.
Qed.

Hint Rewrite <- @fold_homomorphism : homomorphism. 
Hint Rewrite <- @fold_homomorphism_2 : homomorphism.
Hint Rewrite <- @fold_homomorphism_3 : homomorphism.
Hint Rewrite  @left_neutral  : homomorphism.
Hint Rewrite  @right_neutral  : homomorphism.
Hint Rewrite <- @Homomorphism'_list_destruct : homomorphism.