Require Import Coq.Lists.List.
Require Import Coq.Arith.Min.
Require Import Arith.

Require Import LIFO.List.Rev.
Require Import LIFO.List.Firstn_skipn.
Require Import LIFO.List.Map.
Require Import LIFO.List.Notations.


  (** ** Inits and Tails *)
Section initstails.
Set Implicit Arguments.
  Variable A: Type.

  (** *** Tails 
      compute the lists of suffixes 
      - example : tails [1,2,3,4]=[[2,3,4], [3,4], [4], []] 
     *)
  Fixpoint tails (l:list A) :=
    match l with
      | nil => nil
      | a::r => r::(tails r)
    end.
  
  (** tails [l] has same length as [l] *)
  Lemma tails_length (l:list A) :
    length (tails l) = length l.
  Proof.
    intros.
    induction l.
    simpl. reflexivity.
    simpl. f_equal. assumption.
  Qed.


  (** [n]th element of [tails l] is  [skipn (n+1) l] *)
  Lemma tails_nth (l:list A) (n:nat) :
    nth n (tails l) nil = skipn (S n) l.
  Proof.
    intros. generalize l.
    induction n.
    intro l'. destruct l'. simpl. reflexivity. simpl. reflexivity.
    intro l'. destruct l'. simpl. reflexivity.
    simpl nth. change (skipn (S (S n)) (a::l')) with (skipn (S n) l').
    apply IHn.
  Qed.

  
  (** *** Inits 
     compute the lists of prefixes 
     - example : [inits  [1,2,3,4]=[[], [1], [1,2], [1,2,3]]] 
     *)
  Definition inits (l:list A) :=
    rev' (map (rev'(A:=A)) (tails (rev'(A:=A) l))).

  (** [inits l] has same length as [l] *)
  Lemma inits_length (l:list A):
    length (inits l) = length l.
  Proof.
    induction l. auto.
    unfold inits.
    rev'_to_rev.
    rewrite rev_length.
    rewrite map_length.
    rewrite tails_length.
    rewrite rev_length.
    reflexivity.
  Qed.

  (** [n]th element of [inits l] is  [firstn n l] *)
  Lemma inits_nth (l:list A) (n:nat) :
    nth n (inits l) l = firstn n l.
  Proof.
    intros.
    assert (n < length l \/ length l <= n) by omega. destruct H.
    unfold inits.
    rev'_to_rev.
    rewrite rev_nth.
    replace l with (rev (rev l)). (* by rev_involutive. *)
    rewrite map_nth. repeat (rewrite rev_involutive).
    rewrite nth_indep with (d':=nil(A:=A)).
    rewrite tails_nth.
    rewrite minus_Sn_m.
    simpl.
    rewrite map_length. rewrite tails_length.
    rev_to_rev'.
    rewrite <- firstn_rev'.
    rev'_to_rev.
    rewrite rev_involutive. reflexivity.
    rev'_to_rev. 
    rewrite map_length. rewrite tails_length. rewrite rev_length. omega.
    rewrite map_length. rewrite tails_length. rewrite rev_length. omega.
    apply rev_involutive. rewrite map_length. rewrite tails_length. rewrite rev_length. assumption.

    rewrite nth_overflow.
    generalize l H. clear l H. induction n.
    intro. destruct l. auto.
    intros. simpl in H. absurd (S (length l) <= 0). intuition. assumption.
    intros. simpl.
    destruct l. reflexivity. f_equal. apply IHn.
    simpl in H. omega.
    rewrite inits_length. assumption.
  Qed.

  (** inits_id_tails : list of (prefix, (id, suffix)) of l
     example :
     [inits_id_tails [1,2,3,4]=[ ([],1,[2,3,4]), ([1],2,[3,4]), ([1,2],3,[4]),([1,2,3],4,[])]]
     *)
  Definition inits_id_tails (l:list A) :=
    combine (inits l) (combine l (tails l)).


  (** [inits_id_tails l] has same length as [l] *)
  Lemma inits_id_tails_length (l:list A) :
    length (inits_id_tails l) = length l.
  Proof.
    intros.
    unfold inits_id_tails.
    rewrite combine_length.
    rewrite inits_length.
    rewrite combine_length.
    rewrite tails_length.
    assert (forall x, Min.min x x = x).
    intros. apply Min.min_l. auto.
    rewrite H.
    apply H.
  Qed.
  
  (** decompTails :  [tails l] decomposition with append *)
  Lemma decompTails : forall (x  y : list A), (tails  (x++y)) = (map (fun z=> z++y) (tails x)) ++ (tails y).
  Proof.
    induction x.
    simpl.
    intro.
    reflexivity.
    intros y.
    simpl.
    rewrite (IHx y).
    reflexivity.
  Qed.
End initstails.

Example tails_informal : tails [1;2;3;4]=[[2;3;4]; [3;4]; [4]; [] ] . reflexivity. Qed.
Example inits_informal: inits [1;2;3;4]= [[]; [1]; [1; 2]; [1; 2; 3]]. reflexivity. Qed.
Example inits_id_tails_informal : inits_id_tails [1;2;3;4]= [([], (1, [2; 3; 4])); ([1], (2, [3; 4])); ([1; 2], (3, [4])); ([1; 2; 3], (4, []))] . reflexivity. Qed.

(** *alternate inits and tails definition, proof of equivalence and some properties.*)
Section alternate_inits_definition.
Set Implicit Arguments.
  Variable A:Type.

(** alternate inits definition.*)  
  Fixpoint inits2 (l : list A) := 
    match l with 
      | nil => nil
      | h::t=> nil :: map (fun t=>h::t) (inits2 t)
    end.

(** alternate tails definition.*)
  Lemma tails_app_end :
    forall (l : list A) (a : A),
      tails(l ++ a::nil) =
      (map (fun l=>l++ a::nil) (tails l)) ++ (nil::nil).
  Proof.
    induction l.
    auto.
    simpl.
    intros a0.
    rewrite <- (IHl a0).
    reflexivity.
  Qed.

(** [map] with function extensional egality, WARNING a less general version exists in [list_tools].*)  
  Definition  map_eq_fun  := map_ext .
  (*   forall (A B: Set) (l : list A) (f g :A->B), *)
  (*     (forall(x:A),f x = g x)-> map f l = map g l. *)
  (* Proof. *)
  (*   intros A0 B l f g H. *)
  (*   induction l. *)
  (*   (* Base case *) *)
  (*   auto. *)
  (*   (* Inductive case *) *)
  (*   simpl. *)
  (*   rewrite H.  *)
  (*   rewrite IHl. *)
  (*   reflexivity. *)
  (* Qed. *)

(* Require Import List_tools. *)
(** equivalence of inits definitions.*)
  Lemma eq_inits :
    forall (l : list A), inits l = inits2 l.
  Proof.
    induction l.
    auto.
    simpl.
    unfold inits.
    assert(forall (t:list A) (h:A), rev'(h::t)=(rev' t)++(h::nil)) as H by(setoid_rev'_to_rev ; auto).
    rewrite H.
    rewrite tails_app_end.
    rewrite map_app. simpl.
    rewrite map_map.
    rev'_to_rev.
    rewrite rev_unit.
    rewrite <- map_rev.
    set(f:=(fun x : list A => rev' (x ++ a :: nil))).
    set(g:= fun l => (@cons A a)  (rev'(A:=A) l)).
    assert(forall x, (f x)=(g x)) as Hf.
    intros x; unfold f;unfold g;  rev'_to_rev; rewrite rev_unit;reflexivity.
    rewrite <- IHl.
    unfold inits.
    rev'_to_rev.
    rewrite (map_eq_fun  f g Hf).
    unfold g.
     rewrite <- map_map.
    rewrite map_rev.
    rev'_to_rev.
    reflexivity.
  Qed.

(** Decomposition of inits over a concatenation of lists .*)  
Lemma decompInits : 
  forall (x  y : list A), 
    inits  (x++y) = (inits x) ++ (map (fun z=> x++z) (inits y)).
Proof.
  intros x y.
  rewrite eq_inits. 
  rewrite eq_inits.
  rewrite eq_inits.
  induction x.
  (* Base case *)
  simpl.
  induction (inits2  y). auto.
  simpl. rewrite <- IHl. reflexivity.
  (* Inductive case *)
  simpl.
  rewrite IHx.
  rewrite map_app.
  set(f:=(fun t : list A => a :: t)).
  set(g:=(fun z : list A => x ++ z)).
  assert((fun z : list A => a :: x ++ z)= fun l => f  (g l) ) as H ;auto.
  rewrite H.
  rewrite <- (map_map g f).
  reflexivity.
Qed.

End alternate_inits_definition.


Lemma combine_nth: 
  forall (A B:Type) (l1: list A) (l2:list B) n (d1: A) (d2 : B), 
    n < length l1 -> n < length l2 ->
    nth n (combine l1 l2) (d1,d2) = (nth n l1 d1, nth n l2 d2).
Proof.
  intros A B l1 l2 n d1 d2 Hl1 Hl2.
  generalize dependent l1. 
  generalize dependent l2.
  induction n;
    destruct l1;
      destruct l2; intro Hl1.
  contradict Hl2 ; apply lt_n_O.
  contradict Hl1 ; apply lt_n_O.
  contradict Hl2 ; apply lt_n_O.
  reflexivity.
  contradict Hl2 ; apply lt_n_O.
  contradict Hl1 ; apply lt_n_O.
  contradict Hl2 ; apply lt_n_O.
  simpl. rewrite IHn. 
  reflexivity.
  apply lt_S_n in Hl2. assumption.
  apply lt_S_n in Hl1. assumption.
Qed.
  

(**  length fact  *)
(* Require Import Inits_tails. *)
Hint Rewrite tails_length inits_length inits_id_tails_length : length.
