Require Import Arith.
Require Import NPeano.
Require Import Omega. 
Require Import Program.
Require Import List.
Require Import Bool.

Require Classes.EquivDec.

Require Import LIFO.List.Misc.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Map.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.StrongNth.
Require Import LIFO.Nat.BoundedNat.
Require Import LIFO.Nat.ProofEquality.
Require Import LIFO.List.Flatten.
Require Import LIFO.List.Last.
Require Import LIFO.Option.Option.

Require CertifiedBSML.Primitives.Specification.
Require CertifiedBSML.StandardLibrary.Base.

Require SDPP.Datatypes.Lists.NonEmptyLists.
Require SDPP.Datatypes.Lists.Lists.
Require Import SDPP.CorrectParallelisation.Structure.

Set Implicit Arguments.

Module Type NUMBER. 

  Variable number : Type.
  Variable add : number -> number -> number.
  Variable opp : number -> number.
  Variable sub : number -> number -> number.
  Variable mul : number -> number -> number.
  Variable div : number -> number -> number.

  Notation "x + y" := (add x y) : number_scope.
  Notation "- x" := (opp x) : number_scope.
  Notation "x - y" := (sub x y) : number_scope.
  Notation "x * y" := (mul x y) : number_scope.
  Notation "x / y" := (div x y) : number_scope.

End NUMBER.

Module Type TYPE
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (Number : NUMBER)
  (BsmlPropertiesM : Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(BsmlPropertiesM))
  (BsmlBaseM : StandardLibrary.Base.TYPE(BsmlPrimitives)(BsmlPropertiesM)(ParSigM))
  (Parallel : Parallelisation)
  (ListPart : Lists.TYPE BsmlPrimitives  BsmlPropertiesM ParSigM BsmlBaseM Parallel)
  (NonEmptyListPart : NonEmptyLists.TYPE BsmlPrimitives  BsmlPropertiesM ParSigM BsmlBaseM  Parallel ListPart).

  Import BsmlPrimitives Number BsmlPropertiesM ParSigM BsmlBaseM NonEmptyListPart.

  Open Scope number_scope.
  Definition heatEquationFormula
    (dt dx gamma uI uIMinusOne uIPlusOne : number) : number:=
    gamma * dt / (dx * dx) * ( uIPlusOne + uIMinusOne - uI - uI) + uI.
  Close Scope number_scope.
    
  (* Discretised Heat Equation : 
     u(x,t+1) = (gamma * dt / dx^2) * ( u(x+1,t) + u(x-1,t) - 2 * u(x,t) ) + u(x,t)
  *)

  Definition heatEquationStepSpecification 
    (step: number -> number -> number -> number -> number -> list number -> list number): 
    Prop :=
    forall (u : list number)(Hu: u<>[])(leftBound rightBound dt dx gamma: number)
      (i : nat)(Hi:i<length u)(default : number),
      nth i (step leftBound rightBound dt dx gamma u) default = 
      heatEquationFormula dt dx gamma 
      (nth i u default) (if beq_nat i 0 then leftBound else nth (i-1) u default) (nth (i+1) u rightBound).

  Fixpoint heatEquationSeq 
    (leftBound rightBound dt dx gamma : number)(u : list number) : list number :=
    match u with 
      | [] => []
      | uI :: u'  => 
        match u' with 
          | [] => [ heatEquationFormula dt dx gamma uI leftBound rightBound ]
          | uIPlusOne:: _ =>
            (heatEquationFormula dt dx gamma uI leftBound uIPlusOne)::
            (heatEquationSeq uI rightBound dt dx gamma u')            
        end
    end.
    
  Lemma heatEquationSeqApp:
    forall (leftBound rightBound dt dx gamma : number)(u1 u2: list number),
      heatEquationSeq leftBound rightBound dt dx gamma (u1 ++ u2) = 
      (heatEquationSeq leftBound (hd rightBound u2) dt dx gamma u1) ++
      (heatEquationSeq (last u1 leftBound) rightBound dt dx gamma u2).
  Proof.
    intros leftBound rightBound dt dx gamma u1;
      revert leftBound rightBound dt dx gamma;
      induction u1 as [ | uI u' IHu1].
        trivial.
        intros leftBound rightBound dt dx gamma u2.
        case_eq(u'++u2); [intros H| intros].
          apply app_eq_nil in H; destruct H as [H1 H2]; subst; trivial.
          simpl.
          rewrite IHu1.
          rewrite H; simpl.
          destruct u' as [ | uIPlusOne u'' Hu'].
            simpl in *; subst; trivial.
            inversion H; subst; rewrite last_indep with (d2:=leftBound); trivial.
  Qed.

  Lemma stepSeqFollowsHeatEquationStepSpecification: 
    heatEquationStepSpecification heatEquationSeq.
  Proof.
    unfold heatEquationStepSpecification.
      induction u as [ | uI u' IHu]; intros Hu leftBound rightBound dt dx gamma i Hi default.
      (* Case nil *)
      contradict Hi; simpl; omega.
      (* Case cons *)
      induction i as [ | i IHi].
        destruct u'; trivial.
        replace (uI::u') with ([uI]++u'); trivial; rewrite heatEquationSeqApp; simpl.
        case_eq u'; intros.
          contradict Hi; subst; simpl; omega.
          subst; rewrite IHu.
            destruct i; simpl.
              trivial.
              replace (i - 0) with i; trivial; omega.
            discriminate.
            apply lt_S_n in Hi; trivial.
  Qed.  

  Notation " p1 == p2 " := (beq_nat p1 p2) (at level 0, no associativity).

  Program Definition projPar (A:Type)(P:A->Prop)(v:par(sig P)) : par A := parfun (@proj1_sig A P) v.

  Program Definition strongHead (A:Type)(l:list A | l<>nil) : 
    { a : A | Some a = head l } := 
    noSome(head l).
  Next Obligation.
    intros A l; destruct l as [l Hl]; simpl; induction l.
      elim Hl; trivial.
      discriminate.
  Defined.

  Definition sHead (A:Type)(l:list A | l<>nil) : A := proj1_sig (strongHead l).

  Lemma sHeadProp:
    forall (A:Type)(l:list A | l<>nil),
      Some( sHead l ) = hd_error (`l).
  Proof.
    intros A l; unfold sHead, strongHead; apply someNoSomeId.
  Qed.

  Fixpoint last_error (A:Type)(l:list A) : option A := 
    match l with 
      | [] => None 
      | [x] => Some x 
      | x::xs => last_error xs
    end.

  Program Definition strongLast (A:Type)(l:list A | l<>nil) : 
    { a : A | Some a = last_error (proj1_sig l) } := noSome(last_error (proj1_sig l)).
  Next Obligation.
    intros A l; destruct l as [l Hl]; simpl; induction l.
      elim Hl; trivial.
      simpl; destruct l.
        discriminate.
        apply IHl; discriminate.
  Defined.

  Definition sLast (A:Type)(l:list A | l<>nil) : A := proj1_sig (strongLast l).

  Program Definition sLastProp:
    forall (A:Type)(l:list A | l<>nil),
      Some (sLast l) = last_error l.
  Proof.
    intros A l; unfold sLast, strongLast; apply someNoSomeId.
  Qed.
       
  Require Import Min.

  Lemma min_lt: 
    forall (n m p:nat), 
      n < p -> min n m < p.
  Proof.
    induction n; intros m p0 H.
      trivial.
      destruct m.
        simpl; omega.
        destruct p0.
          contradict H; omega.
          apply lt_n_S; apply IHn; apply lt_S_n; trivial.
  Qed.

  Program Definition destructProcessor: 
    forall (i:processor),
      i=firstProcessor \/ i=lastProcessor \/ ( 0 < i /\ i < lastProcessor ).
  Proof.
    unfold firstProcessor, lastProcessor; intro i.
      destruct i as [i Hi]; destruct i; simpl.
        left; apply proj1_sigBoundedNatInjective; trivial.
        case_eq(beq_nat (S i) (proj1_sig lastProcessor)); intro Heq. 
           apply beq_nat_true in Heq; subst; right; left; unfold lastProcessor; 
             apply proj1_sigBoundedNatInjective; trivial.
           repeat right; split.
             omega.
             replace (`lastProcessor) with (bsp_p-1) in Heq; auto; destruct bsp_p.
               contradict Hi; omega.
               unfold lt in *;destruct Hi; apply beq_nat_false in Heq.
                 contradict Heq; omega.
                 omega.
  Qed.

  (* A CHANGER: replicate d'une fonction constante *)
  Program Definition constantFunPar (A B:Type)(b:B) : 
    { vr : par (A->B) |  forall i, forall a, (get vr i) a = b } := 
    mkpar(fun pid (a:A) => b).
  Next Obligation.
    intros A B b i a; autorewrite with bsml; trivial.
  Qed.

  Definition lconstantFunPar (A B :Type)(b:B) := 
    proj2_sig(constantFunPar A b).

  Hint Rewrite lconstantFunPar : heat_equation.
                            
  Program Definition 
    getBoundsAux (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) : 
    { vr : par (option A) | forall (i:processor),
      get vr i = Some ( if ( i == firstProcessor ) then leftBound  else sLast (get v (i-1)) ) } *
    { vr : par (option A) | forall (i:processor),
      get vr i = Some ( if ( i == lastProcessor) then rightBound else sHead (get v (min (i+1) lastProcessor))) } :=
    let msg := put(apply(mkpar(fun (pid:processor) data (dst:processor) =>
      if ( dst == (pid+1) ) && negb(pid == (bsp_p-1)) then Some (sLast data)
        else if ( dst == (pid-1) ) && (negb(pid == 0)) then Some (sHead data)
          else None)) (parSig v _ H) ) in
    ( applyat firstProcessor (constantFunPar processor (Some leftBound)) msg (parSig (mkpar(fun pid=>pid-1)) _ _ ) , 
      applyat lastProcessor (constantFunPar processor (Some rightBound)) msg (parSig (mkpar(fun pid=>min (pid+1) lastProcessor)) _ _ ) ).
  Next Obligation.
    intros A leftBound rightBound v H vr i; destruct i; simpl; omega.
  Defined.
  Next Obligation.
    intros; trivial.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H vr i.
    destruct i; destruct lastProcessor; rewrite min_comm; apply min_lt; assumption.
  Defined.
  Next Obligation.
    intros; trivial.
  Defined.
  Next Obligation.
    intros; autorewrite with bsml; destruct i; simpl; omega.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H msg i; 
      autorewrite with bsml bsmlbase par_sig option heat_equation.
    destruct(destructProcessor i) as [Hproc|Hproc]; try destruct Hproc as [Hproc | Hproc]; simpl; subst.
    (* First processor *)
    unfold firstProcessor; subst; trivial.
    (* Last processor *)
    unfold lastProcessor; subst; simpl.
    case_eq(bsp_p == 1); intro Heq. 
      rewrite beq_nat_true_iff in Heq; repeat (rewrite Heq at 1; simpl); trivial. 
      rewrite beq_nat_false_iff in Heq; repeat (rewrite Heq at 1; simpl); trivial.
      destruct bsp_pSMaxPid as [maxPid HmaxPid]; rewrite HmaxPid at 1; simpl.
      rewrite <- minus_n_O.
      destruct maxPid.
        contradict Heq; assumption.
        rewrite HmaxPid at 1; simpl.
        replace (  get (proj1_sig(mkpar (fun pid : processor => `pid - 1)))
          (exist (fun pid : nat => pid < bsp_p) (bsp_p - 1) _ ) + 1 ) with (bsp_p-1).
        rewrite HmaxPid at 1; simpl; rewrite <- beq_nat_refl; simpl.
        rewrite negb_if.
        assert( get (proj1_sig(mkpar (fun pid : processor => `pid - 1)))
          (exist (fun pid : nat => pid < bsp_p) (bsp_p - 1) lastProcessor_obligation_1 ) = bsp_p -1 -1 ) as H1
        by (autorewrite with bsml; simpl; trivial).
        rewrite H1 at 1; simpl; repeat (rewrite HmaxPid at 1); simpl; rewrite <- minus_n_O; simpl.
        replace (maxPid == (S maxPid) ) with false;
        replace ( (bsp_p - 1) == 0) with false; simpl.
        f_equal; simpl;
          apply someInjectiveEq;
            repeat(rewrite sLastProp);
              repeat(f_equal; simpl);
                apply proj1_sigBoundedNatInjective;
                  simpl;
                    autorewrite with bsml;
                      trivial.

        rewrite HmaxPid; symmetry; apply <- beq_nat_false_iff; omega.
        symmetry; apply <- beq_nat_false_iff; omega.
        symmetry; apply <- beq_nat_false_iff; omega.
        autorewrite with bsml; simpl; omega.
    (* Other processor *)
    destruct Hproc as [Hproc1 Hproc2]; destruct i as [i Hi]; destruct i; simpl.
      trivial.
      assert((get (proj1_sig(mkpar (fun pid : processor => `pid - 1)))
            (exist (fun pid : nat => pid < bsp_p) (S i) Hi)) = i) as Heq
      by (autorewrite with bsml; simpl; omega).
      rewrite Heq at 1; rewrite plus_comm; simpl; rewrite <- beq_nat_refl.
      rewrite Heq at 1; simpl; rewrite negb_if.
      replace ( i == (bsp_p - 1) ) with false; simpl.
        f_equal; simpl;
          apply someInjectiveEq;
            repeat(rewrite sLastProp);
              repeat(f_equal; simpl);
                apply proj1_sigBoundedNatInjective;
                  simpl;
                    autorewrite with bsml;
                      trivial.
       symmetry; apply <- beq_nat_false_iff; omega.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H msg i; autorewrite with bsml;
      rewrite min_comm; apply min_lt; unfold lastProcessor; simpl; destruct bsp_pSMaxPid; intuition.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H msg i.
    destruct(destructProcessor i) as [Hproc|Hproc]; try destruct Hproc as [Hproc | Hproc];     
      autorewrite with bsml bsmlbase option par_sig heat_equation; simpl; subst.
    (* First processor *)
    unfold firstProcessor; simpl. 
    case_eq(bsp_p == 1); intro Heq. 
      (* bsp_p = 1 *)
      rewrite beq_nat_true_iff in Heq; repeat(rewrite Heq at 1; simpl); trivial.
      (* bsp_p <> 1 *)
      destruct bsp_pSMaxPid as [maxPid HmaxPid].
      rewrite beq_nat_false_iff in Heq.
      assert(bsp_p - 1 <> 0) as H' by (rewrite HmaxPid; omega).
      pose(H'':=H').
      rewrite <- beq_nat_false_iff in H''; rewrite H'' at 1.
      assert ( get (proj1_sig(mkpar (fun pid : processor => min (`pid + 1) (bsp_p - 1))))
           (exist (fun pid : nat => pid < bsp_p) 0
              firstProcessor_obligation_1) =  1) as H''' by
      (autorewrite with bsml; simpl; destruct (bsp_p - 1); auto; elim H'; trivial).
      repeat(rewrite H''' at 1; simpl).
      f_equal;simpl.
      rewrite HmaxPid at 1; simpl; rewrite <- minus_n_O.
      destruct maxPid. 
        rewrite HmaxPid in Heq; elim Heq; auto.
        apply someInjectiveEq.
        repeat(rewrite sHeadProp).
        repeat(f_equal;simpl).
        apply proj1_sigBoundedNatInjective; simpl.
        autorewrite with bsml; simpl; trivial.
    (* Last processor *)
    unfold lastProcessor; simpl; rewrite <- beq_nat_refl; trivial.
    (* Other processors *)
    destruct bsp_pSMaxPid as [maxPid HmaxPid].
    destruct Hproc as [Hproc1 Hproc2]; destruct i as [i Hi]; destruct i; simpl in *.
      contradict Hproc1; omega.
      replace ( (bsp_p-1) == (S i) ) with false.
      assert((get (proj1_sig(mkpar (fun pid : processor => min (`pid + 1) (bsp_p - 1))))
            (exist (fun pid : nat => pid < bsp_p) (S i) Hi)) = S(S i)) as Heq.
        autorewrite with bsml. simpl; rewrite HmaxPid; simpl; rewrite <- minus_n_O;
        destruct maxPid.  contradict Hproc2; rewrite HmaxPid; omega.
        replace (min (i+1) maxPid) with (S i); auto.
        rewrite plus_comm; simpl; destruct maxPid.
            contradict Hproc2; omega.
            rewrite MinMax.min_l; auto.
            omega.
      rewrite Heq at 1; simpl.
      replace (i == (S (i + 1))) with false; simpl.
      rewrite Heq at 1; simpl; rewrite <- beq_nat_refl; simpl; rewrite negb_if.
      rewrite Heq at 1; simpl.
      rewrite HmaxPid at 1; simpl; rewrite <- minus_n_O. destruct maxPid.
            contradict Hproc2; omega.
      replace ( i == maxPid ) with false; simpl.
        f_equal; simpl;
          apply someInjectiveEq;
            repeat(rewrite sHeadProp);
              repeat(f_equal; simpl);
                apply proj1_sigBoundedNatInjective;
                  simpl;
                    autorewrite with bsml;
                      trivial.
       symmetry; apply <- beq_nat_false_iff; omega.
       symmetry; apply <- beq_nat_false_iff; omega.       
       symmetry; apply <- beq_nat_false_iff; omega.
  Defined.

  Definition lgetBoundsAuxLeft (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) :=
    proj2_sig(fst(getBoundsAux leftBound rightBound H)).

  Definition lgetBoundsAuxRight (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) :=
    proj2_sig(snd(getBoundsAux leftBound rightBound H)).

  Hint Rewrite lgetBoundsAuxLeft lgetBoundsAuxRight : heat_equation.

  Obligation Tactic := idtac.

  Program Definition 
    getBounds (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) : 
    { vr : par A | forall (i:processor),
      get vr i = if ( i == firstProcessor ) then leftBound  else sLast (get v (i-1))  } *
    { vr : par A | forall (i:processor),
      get vr i = if ( i == lastProcessor) then rightBound else sHead (get v (min (i+1) lastProcessor)) } :=
    let tmp := getBoundsAux leftBound rightBound H in
    ( parfun (@noSome A) (parSig (fst tmp) _ _) , 
      parfun (@noSome A) (parSig (snd tmp) _ _) ).
  Next Obligation.
    program_simpl; destruct i; simpl; omega.
  Defined.
  Next Obligation.
    program_simpl.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H vr i.
    destruct i; destruct lastProcessor; rewrite min_comm; apply min_lt; assumption.
  Defined.
  Next Obligation.
    intros; trivial.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H tmp i;
      replace (fst (proj1_sig(fst tmp), snd tmp)) with (proj1_sig(fst tmp));
        autorewrite with bsmlbase option par_sig heat_equation; auto; discriminate.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H tmp i.
    rewrite_parfun; rewrite lparSig; apply someInjectiveEq; rewrite someNoSomeId.
    replace (get (fst (proj1_sig(fst tmp), snd tmp)) i) with (get (proj1_sig(fst tmp)) i); auto.
    autorewrite with bsmlbase option par_sig heat_equation.
    f_equal; destruct( (`i) == (`firstProcessor) ); trivial.
(*  apply someInjectiveEq;
      repeat(rewrite sLastProp);
        f_equal; simpl; f_equal; simpl;
          apply proj1_sigBoundedNatInjective; simpl; auto. *)
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H tmp i;
      replace (snd ((fst tmp), proj1_sig(snd tmp))) with (proj1_sig(snd tmp));
        autorewrite with bsmlbase option par_sig heat_equation; auto; discriminate.
  Defined.
  Next Obligation.
    intros A leftBound rightBound v H tmp i.
    bsmlbase; apply someInjectiveEq; autorewrite with bsmlbase heat_equation par_sig option.
    replace (get (snd ( (fst tmp), (proj1_sig(snd tmp)) ) ) i)  with (get (proj1_sig(snd tmp)) i).
    autorewrite with bsmlbase heat_equation par_sig option.
    f_equal;
    destruct( (`i) == (`firstProcessor) ); trivial.
    apply someInjectiveEq;
      repeat(rewrite sLastProp);
        f_equal.
  Defined.

  Definition lgetBoundsLeft (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) :=
    proj2_sig(fst(getBounds leftBound rightBound H)).

  Definition lgetBoundsRight (A:Type)(leftBound rightBound : A)(v: par(list A))(H: forall i, get v i <> nil) :=
    proj2_sig(snd(getBounds leftBound rightBound H)).

  Hint Rewrite lgetBoundsLeft lgetBoundsRight : heat_equation.

  Program Definition heatEquationPar 
    (leftBound rightBound dt dx gamma : number)(u : par(list number))(Hu: forall i, get u i <> nil) : par(list number) :=
    let bounds := getBounds leftBound rightBound Hu in 
      apply (parfun2 (fun l r => heatEquationSeq l r dt dx gamma) (fst bounds) (snd bounds)) u.

  Program Definition rightBoundAtI 
    (rightBound : number)(u : par(list number))(Hu: forall i, get u i <> nil)(i:processor) : 
    number :=
    if ( (`i) == (`lastProcessor) ) 
      then rightBound 
      else sHead (get u (min (i+1) lastProcessor)).
  Next Obligation.
    intros rightBound u Hu i;
      unfold lastProcessor; destruct i; simpl; rewrite min_comm; apply min_lt; omega.
  Defined.
  Next Obligation.
    auto.
  Defined.

  Program Definition leftBoundAtI 
    (leftBound : number)(u : par(list number))(Hu: forall i, get u i <> nil)(i:processor) : 
    number :=
    if ( (`i) == (`firstProcessor) ) then leftBound  else sLast (get u (i-1)).
  Next Obligation.
    program_simpl; destruct i; simpl; omega.
  Defined.
  Next Obligation.
    program_simpl.
  Defined.

  Program Definition heatParSeqEqAtI: 
    forall (leftBound rightBound dt dx gamma : number)(u : par(list number))(Hu: forall i, get u i <> nil)(i:processor),
      (get (heatEquationPar leftBound rightBound dt dx gamma Hu) i) = 
      heatEquationSeq (leftBoundAtI leftBound Hu i) (rightBoundAtI rightBound Hu i) 
      dt dx gamma (get u i).
  Proof.
    intros leftBound rightBound dt dx gamma u Hu i;
      unfold heatEquationPar, leftBoundAtI, rightBoundAtI;
        set(bounds:=getBounds leftBound rightBound Hu);
          replace ( fst (proj1_sig(fst bounds), snd bounds) ) with (proj1_sig (fst bounds)); auto;
            replace ( snd (fst bounds, proj1_sig(snd bounds) ) ) with (proj1_sig (snd bounds)); auto;
              unfold bounds;
                autorewrite with bsmlbase option par_sig;
                  rewrite lapply;
                    rewrite lparfun2 with(f:=(fun l r : number => heatEquationSeq l r dt dx gamma));
                      autorewrite with bsmlbase option par_sig heat_equation; auto.
    f_equal.
    (*    destruct(  (`i) == (`firstProcessor) ); trivial. 
    rewrite someInjectiveEq.
    repeat(rewrite sLastProp).
    f_equal. 
    autorewrite with bsmlbase heat_equation option par_sig.
    replace (proj1_sig(exist  (get u (exist (fun pid : nat => pid < bsp_p) ((proj1_sig i) - 1)
             (getBounds_obligation_1 leftBound rightBound Hu
                (proj1_sig(fst (getBounds leftBound rightBound Hu))) i)))
       (getBounds_obligation_2 leftBound rightBound Hu
          (proj1_sig(fst (getBounds leftBound rightBound Hu))) i)) ) with (get u (exist (fun pid : nat => pid < bsp_p) ((proj1_sig i) - 1)
             (getBounds_obligation_1 leftBound rightBound Hu
                (proj1_sig(fst (getBounds leftBound rightBound Hu))) i))).
    replace (proj1_sig(exist (fun l : list number => l <> [])
       (get u
          (exist (fun pid : nat => pid < bsp_p) ((proj1_sig i) - 1)
             (leftBoundAtI_obligation_1 leftBound Hu i)))
       (leftBoundAtI_obligation_2 leftBound Hu i))) with 
    (get u
      (exist (fun pid : nat => pid < bsp_p) ((proj1_sig i) - 1)
        (leftBoundAtI_obligation_1 leftBound Hu i))).
    f_equal; apply proj1_sigBoundedNatInjective; simpl; trivial.
    auto.
    auto. *)
    destruct(  (`i) == (`lastProcessor) ); trivial;
      rewrite someInjectiveEq; 
        repeat(rewrite sHeadProp); f_equal; simpl; f_equal; 
          apply proj1_sigBoundedNatInjective; trivial.
  Qed.

  Lemma hdHd_error: 
    forall (A:Type)(l:list A), l<>nil -> forall a:A, Some(hd a l) = hd_error l.
  Proof.
    induction l as [ | head tail IH]; intros H a.
      elim H; trivial.
      trivial.
  Qed.

  Lemma lastAppCons: 
    forall (A:Type)(l:list A)(a d:A),
      last (l++a::nil) d  = a.          
  Proof.
    induction l.
    trivial.
    intros a0 d.
    case_eq(l++[a0]);intros.
    destruct l; simpl in H; discriminate.
    simpl; rewrite H; rewrite <- H; apply IHl.
  Qed.

  Lemma last_errorAppCons: 
    forall (A:Type)(l:list A)(a:A),
      last_error (l++a::nil) = Some a.
  Proof.
    intros A l a.
    induction l.
      trivial.
      case_eq(l++[a]);intros.
        destruct l; simpl in H; discriminate.
        simpl; rewrite H; rewrite <- H; assumption.
  Qed.
  
  Lemma last_errorApp: 
    forall (A:Type)(l1 l2:list A),
      l2<>nil -> last_error (l1++l2) = last_error l2.
  Proof.
    induction l1 using rev_ind.
      trivial.
      intros; rewrite <- app_assoc; rewrite IHl1. 
        simpl; destruct l2.
          elim H; trivial.
          trivial.
        discriminate.
  Qed.
  
  
  Lemma lastLast_error: 
    forall (A:Type)(l:list A), 
      l<>nil -> (forall a:A, Some (last l a) = last_error l).
  Proof.
    induction l using rev_ind; intros H default.
      elim H; trivial.
      rewrite last_errorAppCons; rewrite lastAppCons; trivial.
  Qed.


  Theorem heatEquationParSeqEq : 
    forall (leftBound rightBound dt dx gamma : number)
      (u : par(list number))(Hu: forall i, get u i <> nil),
      heatEquationSeq leftBound rightBound dt dx gamma (listOfParList u) = 
      listOfParList(heatEquationPar leftBound rightBound dt dx gamma Hu). 
  Proof.
    intros leftBound rightBound dt dx gamma u Hu.
    (* We first need to replace right bound by (rightBoundAtI rightBound Hu lastProcessor)
       in the right hand side so that it is generalized by the inductin tactique
       *)
    replace (heatEquationSeq leftBound rightBound dt dx gamma (listOfParList u)) with
      (heatEquationSeq leftBound (rightBoundAtI rightBound Hu lastProcessor) dt dx gamma (listOfParList u)) by (
    f_equal; unfold rightBoundAtI;
      replace ( (`lastProcessor) == (`lastProcessor) ) with true by( apply beq_nat_refl);
        reflexivity) .

    listOfParListInduction.
        rewrite heatParSeqEqAtI.
         f_equal.

      rewrite heatEquationSeqApp.
      f_equal.
        (* First part *)
        rewrite <- IHp'. f_equal. unfold rightBoundAtI.
        apply someInjectiveEq.
        rewrite (hdHd_error (Hu S_p)).
        replace ( (proj1_sig p') == (proj1_sig lastProcessor)) with false by (
          simpl;symmetry; rewrite beq_nat_false_iff; omega).       
        rewrite sHeadProp. 
        cbv beta iota delta [proj1_sig].
        repeat f_equal. apply proj1_sigBoundedNatInjective; simpl; rewrite min_l; omega.
        (* Second part *)
        rewrite heatParSeqEqAtI; f_equal.
        unfold leftBoundAtI.
        replace ((proj1_sig S_p) == (`firstProcessor)) 
          with false by auto.
        apply someInjectiveEq.
        rewrite sLastProp.
        rewrite lastLast_error.
        cbv beta iota delta [proj1_sig].
        case_eq( p  == (proj1_sig firstProcessor)); intro H.
          rewrite beq_nat_true_iff in H;
            replace (processorsUpTo p' ) with ([firstProcessor])
            by (replace  p' with firstProcessor 
              by (apply proj1_sigBoundedNatInjective; simpl; unfold firstProcessor in H; 
                simpl in H; rewrite H; trivial);
            unfold processorsUpTo; simpl; f_equal; apply proj1_sigBoundedNatInjective; trivial). 
            replace (exist (fun pid : nat => pid < bsp_p)
              ((let (a, _) := S_p in a) - 1)
              (leftBoundAtI_obligation_1 leftBound Hu S_p)) with firstProcessor 
            by (apply proj1_sigBoundedNatInjective; simpl; unfold firstProcessor in H; 
              simpl in H; rewrite H; trivial);
            simpl; rewrite app_nil_r; trivial.
          rewrite beq_nat_false_iff in H.
            unfold firstProcessor in H; simpl in H; destruct p.
              elim H; trivial.
              subst p'.
              rewrite processorsUpToApp.
              rewrite map_app.
              rewrite flattenApp.
              rewrite last_errorApp.
              Opaque leftBoundAtI_obligation_1.
              simpl.
              rewrite app_nil_r.
              f_equal; f_equal.
              apply proj1_sigBoundedNatInjective; trivial.
              simpl; rewrite app_nil_r; apply Hu.
              apply flattenNotNil.
                intros l Hl. apply in_map_iff in Hl; 
                  destruct Hl as [proc Hl]; destruct Hl as [Hproc Hl];
                    rewrite <- Hproc; apply Hu.
                unfold processorsUpTo; simpl; discriminate.
  Qed.

Lemma  heatEquationSeqLength :  forall u leftBound rightBound dt dx gamma, length (heatEquationSeq leftBound rightBound dt dx gamma u ) = length u.
induction u.
reflexivity.
simpl.
destruct u.
reflexivity.
intros leftBound rightBound dt dx gamma.
Opaque heatEquationSeq.
simpl;rewrite IHu; reflexivity.
Qed.
Hint Rewrite heatEquationSeqLength : length.

Program Definition heatEquationSeqRestricted
  (leftBound rightBound dt dx gamma : number)(u : {l: list number | length l >= bsp_p}) :  {l: list number | length l >= bsp_p}:=
  heatEquationSeq leftBound rightBound dt dx gamma u.

Next Obligation.
  intros leftBound rightBound dt dx gamma u.
  destruct u as [u u_spec] ;simpl.
  rewrite heatEquationSeqLength.
  assumption.
Qed.

Program Definition heatEquationParRestricted (leftBound rightBound dt dx gamma : number)(pu : {plst: par (list number) |  forall i, get plst i<>nil}) :  {plst: par (list number) |  forall i, get plst i<>nil}:=
  heatEquationPar leftBound rightBound dt dx gamma   (proj2_sig pu).
Next Obligation.
  intros leftBound rightBound dt dx gamma pu i.
  rewrite heatParSeqEqAtI.
  apply  Length.lengthPosNotNil.
  rewrite  heatEquationSeqLength.
  destruct pu.
  generalize  n i;simpl;
  auto with list.
Qed.
Import Parallel.
Instance heatEquationParallel (leftBound rightBound dt dx gamma : number) :  
                     Parallel 
                     (heatEquationSeqRestricted leftBound rightBound dt dx gamma)
                     (heatEquationParRestricted leftBound rightBound dt dx gamma).
constructor.
intros ap. simpl in *.
destruct ap.
unfold heatEquationParRestricted, heatEquationSeqRestricted, listOfNonEmptyParList.
apply SeqListGeBspp_proofIrrel.
simpl.
symmetry; apply  heatEquationParSeqEq.
Qed.

End TYPE.

Module Make
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (Number : NUMBER)
  (BsmlPropertiesM : Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(BsmlPropertiesM))
  (BsmlBaseM : StandardLibrary.Base.TYPE(BsmlPrimitives)(BsmlPropertiesM)(ParSigM))
  (Parallel : Parallelisation)
  (ListPart : Lists.TYPE BsmlPrimitives  BsmlPropertiesM ParSigM BsmlBaseM Parallel)
  (NonEmptyListPart : NonEmptyLists.TYPE BsmlPrimitives  BsmlPropertiesM ParSigM BsmlBaseM  Parallel ListPart).

  Include TYPE(BsmlPrimitives)(Number)(BsmlPropertiesM)(ParSigM)(BsmlBaseM) Parallel ListPart NonEmptyListPart.

End Make.  
