(** *Heat Equation using BH  *)
Unset Automatic Introduction.    (* coq 8.3 *)

Require Import Arith.
Require Import NPeano.
Require Import Omega. 
Require Import Program.
Require Import List.
Require Import Bool.

Require Export Skel.BH.bh.
Require Export Skel.BH.bh_bsml.
Require Export Skel.BH.bh_classes.

Require Export CertifiedBSML.QuickStart.

Require Import LIFO.List.Permute.
Require Classes.EquivDec.
Require Import LIFO.List.Misc.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Map.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.StrongNth.
Require Import LIFO.Nat.BoundedNat.
Require Import LIFO.Nat.ProofEquality.
Require Import LIFO.List.Flatten.
Require Import LIFO.List.Last.
Require Import LIFO.List.Head.
Require Import LIFO.Option.Option.
Require Import LIFO.Algebra.AlgebraClasses.

Module Type NUMBER. 

  Variable number : Type.
  Variable add : number -> number -> number.
  Variable opp : number -> number.
  Variable sub : number -> number -> number.
  Variable mul : number -> number -> number.
  Variable div : number -> number -> number.
  Notation "x + y" := (add x y) : number_scope.
  Notation "- x" := (opp x) : number_scope.
  Notation "x - y" := (sub x y) : number_scope.
  Notation "x * y" := (mul x y) : number_scope.
  Notation "x / y" := (div x y) : number_scope.

End NUMBER.

Module TYPE
  (Number : NUMBER)
  (Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh).
  
Import Bh_Classes.  
  Import Bsml_bh.  
  Import Parallel.


Set Implicit Arguments.

Import Number.

  Open Scope number_scope.
  (** Discretised Heat Equation : 
     u(x,t+1) = (gamma * dt / dx^2) * ( u(x+1,t) + u(x-1,t) - 2 * u(x,t) ) + u(x,t)
     *)
  Definition heatEquationFormula
    (dt dx gamma uI uIMinusOne uIPlusOne : number) : number:=
    gamma * dt / (dx * dx) * ( uIPlusOne + uIMinusOne - uI - uI) + uI.
  Close Scope number_scope.
  
  (**  A computation step for HeatEquation has to follow this specification   *)
  Definition heatEquationStepSpecification 
    (step: number -> number -> number -> number -> number -> list number -> list number): 
    Prop :=
    forall (u : list number)(Hu: u<>[])(leftBound rightBound dt dx gamma: number)
      (i : nat)(Hi:i<length u)(default : number),
      nth i (step leftBound rightBound dt dx gamma u) default = 
      heatEquationFormula dt dx gamma 
      (nth i u default) (if beq_nat i 0 then leftBound else nth (i-1) u default) (nth (i+1) u rightBound).

  (**  Sequential implementation of the equation computation   *)
  Fixpoint heatEquationSeq 
    (leftBound rightBound dt dx gamma : number)(u : list number) : list number :=
    match u with 
      | [] => []
      | uI :: u'  => 
        match u' with 
          | [] => [ heatEquationFormula dt dx gamma uI leftBound rightBound ]
          | uIPlusOne:: _ =>
            (heatEquationFormula dt dx gamma uI leftBound uIPlusOne)::
            (heatEquationSeq uI rightBound dt dx gamma u')            
        end
    end.
    
  (**  Sequential implementation decomposition over list concatenation   *)
  Lemma heatEquationSeqApp:
    forall (leftBound rightBound dt dx gamma : number)(u1 u2: list number),
      heatEquationSeq leftBound rightBound dt dx gamma (u1 ++ u2) = 
      (heatEquationSeq leftBound (hd rightBound u2) dt dx gamma u1) ++
      (heatEquationSeq (last u1 leftBound) rightBound dt dx gamma u2).
  Proof.
    intros leftBound rightBound dt dx gamma u1;
      revert leftBound rightBound dt dx gamma;
      induction u1 as [ | uI u' IHu1].
        trivial.
        intros leftBound rightBound dt dx gamma u2.
        case_eq(u'++u2); [intros H| intros].
          apply app_eq_nil in H; destruct H as [H1 H2]; subst; trivial.
          simpl.
          rewrite IHu1.
          rewrite H; simpl.
          destruct u' as [ | uIPlusOne u'' Hu'].
            simpl in *; subst; trivial.
            inversion H; subst; rewrite last_indep with (d2:=leftBound); trivial.
  Qed.

  (**  Sequential implementation follow specification   *)
  Lemma stepSeqFollowsHeatEquationStepSpecification: 
    heatEquationStepSpecification heatEquationSeq.
  Proof.
    unfold heatEquationStepSpecification.
      induction u as [ | uI u' IHu]; intros Hu leftBound rightBound dt dx gamma i Hi default.
      (* Case nil *)
      contradict Hi; simpl; omega.
      (* Case cons *)
      induction i as [ | i IHi].
        destruct u'; trivial.
        replace (uI::u') with ([uI]++u'); trivial; rewrite heatEquationSeqApp; simpl.
        case_eq u'; intros.
          contradict Hi; subst; simpl; omega.
          subst; rewrite IHu.
            destruct i; simpl.
              trivial.
              replace (i - 0) with i; trivial; omega.
            discriminate.
            apply lt_S_n in Hi; trivial.
  Qed.  
Require Import LIFO.List.Main.
Require  Import ProgramCalculation.Tactics.


(**  Sequential implementation encapsulation for parallelisation with BH   *)
Definition heatEquationSeqCompat  (gblLeftBound gblRightBound : number) 
  (leftBound rightBound :option number) dt dx gamma u:=
  heatEquationSeq (no_some_default leftBound gblLeftBound) (no_some_default rightBound gblRightBound) dt dx gamma u.


(**  Link between both sequential implementation   *)
 Lemma  heatEquationSeqCompat_heatSeq  (gblLeftBound gblRightBound : number) 
   (leftBound rightBound :option number) (dt dx gamma : number)(u: list number):
   heatEquationSeqCompat gblLeftBound gblRightBound leftBound rightBound dt dx gamma u
   =
   heatEquationSeq 
   (match leftBound with None =>gblLeftBound | Some a => a end )
   (match rightBound with None =>gblRightBound | Some a => a end)
     dt dx gamma u
   .
Proof.
  Begin.
    LHS    
  ={unfold heatEquationSeqCompat}
    (heatEquationSeq (no_some_default leftBound gblLeftBound) (no_some_default rightBound gblRightBound) dt dx gamma u). 
 ={unfold no_some_default}
    (heatEquationSeq 
   (match leftBound with None =>gblLeftBound | Some a => a end )
   (match rightBound with None =>gblRightBound | Some a => a end)
     dt dx gamma u
     ).
 []. 
Qed.

(**  Encapsulated sequential implementation  decomposition over concatenation  *)
Lemma heatEquationSeqCompatApp   (gblLeftBound gblRightBound : number) :
  forall (leftBound rightBound : option number) (dt dx gamma : number)(u1 u2: list number),
    heatEquationSeqCompat gblLeftBound gblRightBound leftBound rightBound dt dx gamma (u1 ++ u2) = 
    (heatEquationSeqCompat gblLeftBound gblRightBound leftBound (left_option (hd_option u2) rightBound) dt dx gamma u1) ++
    (heatEquationSeqCompat gblLeftBound gblRightBound (right_option  leftBound (last_option u1) ) rightBound dt dx gamma u2).
Proof.    
  intros gblLeftBound gblRightBound leftBound rightBound dt dx gamma u1 u2.
  repeat rewrite  heatEquationSeqCompat_heatSeq. 
  rewrite heatEquationSeqApp.
  f_equal;  destruct leftBound; destruct rightBound;
   (  ( rewrite  hd_hd_option ; progress (destruct (hd_option u2);  reflexivity))
     || 
       (rewrite  last_last_option ;progress (destruct (last_option u1);  reflexivity))
         ||
       reflexivity
       ||
         idtac).
Qed.

(**  Some homomorphism declaration  *)
Instance  right_option_last_option_Homomorphism (A:Type)  : Homomorphism'  (@right_option A) (@last_option A).  
intros A.
constructor;try typeclasses eauto.
 intros x y.
 induction y using rev_ind.
 rewrite app_nil_r; reflexivity.
rewrite last_option_app_r.
replace  (x ++ y ++ [x0]) with  ((x ++ y) ++ [x0]) by apply AlgebraClasses.associative.
rewrite last_option_app_r.
reflexivity.
Defined.

Instance  left_option_hd_option_Homomorphism (A:Type)  :  Homomorphism' (@left_option A) (@hd_option A).
intros A.
   constructor; try typeclasses eauto.
   intros x y.
   case x; reflexivity.
Defined.

 Instance heatEq_bh (default_left default_right dt dx gamma : number) : 
    BH_PROP  (fun u leftBound rightBound  => heatEquationSeqCompat default_left default_right  leftBound rightBound dt dx gamma u).

  Begin.

  check (heatEquationSeqCompat default_left default_right l r dt dx gamma [] = []) is GOAL.
  []. 

  BeginEvar.
  check (heatEquationSeqCompat default_left default_right l r dt dx gamma [a] =   [EVAR  l a r] ) is GOAL.
  LHS
  ={unfold heatEquationSeqCompat}
    ( heatEquationSeq (no_some_default l default_left) (no_some_default r default_right) dt dx gamma [a]).
 ={simpl}
   ( [heatEquationFormula dt dx gamma a (no_some_default l default_left)  (no_some_default r default_right)]).

={simpl}
    ( [(fun l a r => heatEquationFormula dt dx gamma a (no_some_default l default_left)  (no_some_default r default_right)) l a r]).
[]. 
endEvar.

BeginEvar.
check (  
  heatEquationSeqCompat default_left default_right l r dt dx gamma (x ++ y)
  = 
  heatEquationSeqCompat default_left default_right l (EVAR (EVAR0 y) r) dt dx gamma x 
  ++
  heatEquationSeqCompat default_left default_right  (EVAR1 l (EVAR2 x)) r dt dx gamma y
)is GOAL.
LHS
={rewrite heatEquationSeqCompatApp}
( heatEquationSeqCompat default_left default_right l
     (left_option (hd_option y) r) dt dx gamma x
     ++
   heatEquationSeqCompat default_left default_right
     (right_option l (last_option x)) r dt dx gamma y).
[]. 
endEvar.
(** Homomorhisme proofs  *)
typeclasses eauto.  
typeclasses eauto.  
Defined.

Instance heatEqParConstr (leftBound rightBound dt dx gamma : number) : 
  Parallelizable (R:=eq) (fun u => heatEquationSeq leftBound rightBound dt dx gamma u).
Begin.
BeginEvar.
check (  join (EVAR ap) = heatEquationSeq leftBound rightBound dt dx gamma (join ap)) is GOAL.
RHS 
={simpl}
( heatEquationSeqCompat leftBound rightBound (Some leftBound)  (Some rightBound) dt dx gamma (join ap)). 
={simpl}
(( fun  u l r => heatEquationSeqCompat leftBound rightBound l  r dt dx gamma u) (join ap) (Some leftBound) (Some rightBound)  ).
rewrite <-  (spec_match (R:=eq)  (Parallelizable:=(BH_PROP_Parallelizable  _ _ _))).
[]. endEvar.
 Defined.

Definition HeatEqPar (leftBound rightBound dt dx gamma : number) := 
 Eval simpl in  parallel (f := (fun u => heatEquationSeq  leftBound rightBound dt dx gamma u)).

Definition HeatEqPar' := 
 Eval compute [HeatEqPar bh_bsml_opt_comp] in  HeatEqPar.

End TYPE.

Module Make
  (Number : NUMBER)
  (Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh).

  Include TYPE Number Bs Bsml_prop ParSig  Bsml_base Parallel Partition Bsml_bh Bh_Classes. 
End Make.
