(** Definition of Parallel Skeleton for computing Maximum Prefix Sum *)
Unset Automatic Introduction.    (* coq 8.3 *)
Require Import List.
Require Import Bool.

Require Import SDPP.Misc.total_order.
Require Import SDPP.Misc.bounded_set.
Require Import SDPP.Misc.Max_folding.

Require Import SDPP.Skel.map_around.
Require Import SDPP.Skel.scan.
Require Import SDPP.Skel.homomorphism.
Require Import SDPP.Skel.LastPar.

Require Import LIFO.Algebra.AlgebraClasses.
Require Import LIFO.List.Last.

(** Parameters of the Skeleton  *)
Module Type MAXIMUM_PREFIX_SUM_PARAM.
  Variable A :Type.
  Variable plus : A -> A ->A.
  Variable plus_id : A.
  Declare Instance plus_neutral : Neutral plus plus_id.
  Declare Instance plus_assoc : Associative plus.
  Variable R : relation A.
  Variable preoR : PreOrder R.
  Variable totR : TotalOrder (preo:=preoR) R.
  Variable min :A.
  Variable min_smallest : forall a, R min  a.
End MAXIMUM_PREFIX_SUM_PARAM.

(** Module Type describing MPS  *)
Module Type TYPE   (Bs : PRIMITIVES) 
                     (Bsml_prop : Properties.TYPE Bs) 
                     (ParSig : ParSig.TYPE Bs Bsml_prop) 
                     (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
                     (Parallel : Parallelisation )
                     (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
                     (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
                     (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh) 
                     (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes)
                     (LastParallel : LastPar.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition)
                     (Scan : scan.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma)
                     (Homomorphism : homomorphism.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma LastParallel)
                     (mps_p : MAXIMUM_PREFIX_SUM_PARAM).

  Import Bs.
  Import Ma.  
  Import Scan.  
  Import Homomorphism.  
  
  (** Importing Sequential Maximum Prefix Sum, no more used, only used for the definition of Max and an instance of neutral needed for parallelisation *)
  Module Maximum_fold_module:= Maximum_Fold (mps_p).
  Import Maximum_fold_module.
  Import  mps_p.


  (** The current sequential Maximum Prefix Sum implementation  
     fold_left Max get the maximum element of a list,
     it is applied here on the list of prefix sum computed by scan
      
     *)

  Definition mps (l:list A) :=  (fun l => fold_left Max l min)  ((scan A plus plus_id) l).



  Import Parallel.
  Import Partition.
  Import Bh_Classes.


  (** The function [maximum_fold] is a homomorphism so it is parallelizable,    
         The scan function  is parallelizable  (it is  a mapAround, see [scan] library)
         So, the composition is parallelizable         
*)

(** ** The final parallel version of Maximum Prefix Sum *)
Definition mps_bsml := Eval compute 
  [parallel_type
    list_partitionnable 
    parallel 
    parallelComposition  
    comp 
    LastParallel.parallel_last
    (* compo_parallelizable *)
    (* scanp *)
    BHS_Parallelizable
    BHS_parallelisation 
    parallel_parallelizable
    (* BHS_L  BHS_R *)  BHS_k BHS_gl BHS_opl BHS_gr BHS_opr  (* BHS_idl  BHS_idr  *)
    mapAround_to_BH 
    homoPar
    homomorphismParallelKernel]   
  in parallel (f:=mps). 

End TYPE.
Module Maximum_prefix_sum  (Bs : PRIMITIVES) 
                     (Bsml_prop : Properties.TYPE Bs) 
                     (ParSig : ParSig.TYPE Bs Bsml_prop) 
                     (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
                     (Parallel : Parallelisation )
                     (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
                     (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
                     (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh) 
                     (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes)
                     (LastParallel : LastPar.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition)
                     (Scan : scan.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma)
                     (Homomorphism : homomorphism.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma LastParallel)
                     (mps_p : MAXIMUM_PREFIX_SUM_PARAM). 
  Include TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition  Bsml_bh Bh_Classes Ma LastParallel Scan Homomorphism mps_p.
End Maximum_prefix_sum.
