Require Import Classes.RelationClasses.
Require Import Setoid.

Set Implicit Arguments.
Generalizable All Variables.

(** This module provides a framework for Correct Parallelization  *)
Module Type Parallelisation.

(** *parallelization/sequentialization of a type

   join provide a method to "sequentialize" a distributed term

   partition provide a methode to distribute a sequential term

   the sequentialization of the distribution of a term is itself (join :o: partition x = x)

   /!\ if we impose that (partition :o: join = id) , we made huge assumption on data distribution.
 *)
Class Partitionnable (A: Type) := 
{
  parallel_type : Type;
  (* parallel_type_is_parallel : exists B, parallel_type = par B; *)
  join : parallel_type -> A;
  partition : A -> parallel_type;
  join_part_match :  ( forall a : A, join ( partition a ) = a )
}.


Notation "// A \\":=(@parallel_type A _).


Instance pairPartition (A B :Type) {partA : Partitionnable A} {partB : Partitionnable B} : Partitionnable (A*B):={
  parallel_type := (parallel_type (A:=A) *parallel_type (A:=B))%type;
  join := (fun ap_bp => let (ap,bp):= ap_bp in (join ap, join bp));
  partition := (fun a_b => let (a,b):= a_b in (partition a, partition b))
}.
intros a.
destruct a.
rewrite join_part_match.
rewrite join_part_match.
reflexivity.
Defined.

Instance pairPartition_left (A B :Type) {partA : Partitionnable A}  : Partitionnable (A*B)|100 
  :=
  {
  parallel_type := (parallel_type (A:=A) * B)%type;
  join := (fun ap_b => let (ap,b):= ap_b in (join ap,  b));
  partition := (fun a_b => let (a,b):= a_b in (partition a, b))
}.
intros a.
destruct a.
rewrite join_part_match.
reflexivity.
Defined.


Instance pairPartition_right (A B :Type) {partB : Partitionnable B}  : Partitionnable (A*B)|100 
  :=
  {
  parallel_type := (A*parallel_type (A:=B))%type;
  join := (fun a_bp => let (a,bp):= a_bp in (a, join bp));
  partition := (fun a_b => let (a,b):= a_b in ( a, partition b))
}.
intros a.
destruct a.
rewrite join_part_match.
reflexivity.
Defined.

(**  A  function fp it the parallelisation of f if the function apply to the partion of term give a result equivalent to that of applying f (after seqentializing the result)
   # <pre>
         A <-------join-------  par A
          |                       |
          |                       |
          f                      fp
          |                       |
          |                       |
         \/                      \/
         B <------join-------- par B
   </pre>#
%
\begin{center}
  \begin{tikzpicture}[scale=1.5]
    \node (A) at (0,-1) {$A$};
    \node (B) at (1,-1) {$B$};
    \node (Ap) at (0,0) {$A_p$};
    \node (Bp) at (1,0) {$B_p$};
    \draw [-> ] (Ap) to node[above]{$f_p$}(Bp) ; 
    \draw [->] (Bp) to node[right]{$join_B$} (B);
    \draw [->] (Ap) to node[left]{$join_A$} (A);
    \draw [->] (A) to node[above]{$f$}(B); 
  \end{tikzpicture}
\end{center}
%
  *)
Class Parallel `{APart : Partitionnable A} `{BPart : Partitionnable B}  `{eqB : Equivalence B R} (f: A -> B)  (fp : // A \\ ->  // B \\ ) :={ 
  parallel_spec_match : forall ap, R (join ( fp (ap))) (f (join ap))
}.


(**  A  function fp it the parallelisation of f if the function apply to the partion of term give a result equivalent to that of applying f (after seqentializing the result)
   # <pre>
         A <-------join-------  par A
          |                       |
          |                       |
          f                      fp
          |                       |
          |                       |
         \/                      \/
         B <------join-------- par B
   </pre>#
%
\begin{center}
  \begin{tikzpicture}[scale=1.5]
    \node (A) at (0,-1) {$A$};
    \node (B) at (1,-1) {$B$};
    \node (Ap) at (0,0) {$A_p$};
    \node (Bp) at (1,0) {$B_p$};
    \draw [-> ] (Ap) to node[above]{$f_p$}(Bp) ; 
    \draw [->] (Bp) to node[right]{$join_B$} (B);
    \draw [->] (Ap) to node[left]{$join_A$} (A);
    \draw [->] (A) to node[above]{$f$}(B); 
  \end{tikzpicture}
\end{center}
%
  *)

(** A parallelizable specification f has a parallel implementation such that (join :o: parallel_f :o: partition x = f x) *)
(** i.e. the sequentialization of the result of the parallel implementation of f, applied to the parallelization of x is equal to f applied to x *)
(** This class allows to automatically retrieve  the parallel implementation of a function f  *)
Class Parallelizable {A B  :Type } (f: A -> B) `{APart : Partitionnable A} `{BPart :Partitionnable B} `{eqB : Equivalence B R} :={
  parallel :  // A \\ -> // B \\;
  spec_match : forall ap,  R (join ( parallel  ap))   (f (join ap))
}.

Notation "\\ f //" := (@parallel _ _ f).
(** An instance of Parallel always give an instance of Parallelizable
   
   so declaring an instance of [Parallel] enable the automatic retrieving of the parallel function using [Parallelizable] its field [parallel] through this instance.
 *)
Global  
Program Instance parallel_parallelizable   `(f:A ->B)   `{Parallel A B R f fp}: Parallelizable   f:={parallel := fp}.
Next Obligation.
  rewrite <- (@parallel_spec_match A _  _  _ R _  f fp _ ).
  reflexivity.
Qed.

(** Class of functions for which Parallele Composition is a paralleliszation of the sequential composition  *)

Require Import Classes.Morphisms.
Import ProperNotations .
Open Scope signature_scope.

Global Program Instance parallelCompositionCorrecte  `{parf : Parallel B  C f fp (APart:=BPart)(R:=Rc) }   
  `{parg : Parallel  A B g gp (BPart:=BPart) (R:=Rb) }  
  `{morph: Proper _  (Rb ==> Rc) f} 
  : Parallel  (fun a => f (g a))  (fun ap => fp (gp ap)) 
.
Next Obligation.
  firstorder.  
  rewrite parallel_spec_match0.
  rewrite parallel_spec_match1.
  reflexivity.  
Qed.

Definition comp (A B C:Type) (f:B->C) (g:A->B) :=(fun (a:A)=> f ( g a)).
Implicit Arguments comp [A B C].

Global Notation " f :o: g" := (comp f g) (at level 100, right associativity) .
Global Program Instance parallelCompositionCorrect'  `{parf : Parallel B  C f fp (APart:=BPart)(R:=Rc) }   
  `{parg : Parallel  A B g gp (BPart:=BPart) (R:=Rb) }  
  `{morph: Proper _  (Rb ==> Rc) f} 
  : Parallel  (f :o: g)  (fp :o: gp) 
:=parallelCompositionCorrecte.

(** Class of functions for which Parallele Composition is a paralleliszation of the sequential composition  *)
Set Printing All.
Global Program Instance parallelComposition  `{parf : Parallelizable B  C f (APart:=BPart)(R:=Rc) }   
  `{parg : Parallelizable  A B g (BPart:=BPart) (R:=Rb) }  
    `{morph: Proper _  (Rb ==> Rc) f}
  : Parallelizable  (fun a => f (g a))
  :={
    parallel:=
    fun (ap: parallel_type (A:=A)) => (parallel  (Parallelizable:=parf)  (f:=f)) ((parallel (Parallelizable:=parg)  (f:=g)) ap)
  }.
Next Obligation.
  rewrite spec_match.
  rewrite (morph  (@join B BPart
              (@parallel A B g APart BPart Rb eqB0 parg ap))(g (@join A APart ap))) .
  reflexivity.  
  rewrite spec_match.
  reflexivity.
Qed.
  

Hint Unfold parallelCompositionCorrecte parallelComposition  : Parallel.
Hint Unfold parallel_parallelizable : Parallel.
Hint Unfold parallel : Parallel.

End Parallelisation.

Module Make .
  Include  (Parallelisation).
End Make.