Require Import  Equalities.
Require Import  CertifiedBSML.Primitives.Specification.



(** *Representation of the distribution of a structure of elements among processors. *)
Module Type DISTRIBUTION_NO_INDEX  (Bs: PRIMITIVES ).
  Import Bs.

  Parameter T:Type.
  (** The sequetial view of the distributed data type *)
  Parameter t : T.

(** As the distribution may be something heavy to compute, we could need a datatype to store it. *)
  Parameter distribution : Type.
  
 End DISTRIBUTION_NO_INDEX.



(** *Representation of the distribution of a structure of elements among processors. 
    It enables to optionally give, for any index    
    the processor [p] where the corresponding element is stored. 
   
 *)
Module Type DISTRIBUTION_OPTION  (Bs: PRIMITIVES ) <: DISTRIBUTION_NO_INDEX Bs.
  Import Bs.

  Parameter T:Type.
  (** The sequetial view of the distributed data type *)
  Parameter t : T.

  (** The type for indexing element in the sequetial view of the distributed data type *)
  Parameter index : Type.

(** As the distribution may be something heavy to compute, we could need a datatype to store it. *)
  Parameter distribution : Type.
  
(** Using a distribution we can compute the location of an element, if it exists  *)
  Parameter get_proc : distribution -> index -> option processor.
 End DISTRIBUTION_OPTION.


(** *Representation of the distribution of a structure of elements among processors.
   It enables to give, for a valid index of an element in the sequential representation,
   the processor [p] where the element is stored. 
   
   This version need a proof that the given index is in the distribution
   
*)
Module Type DISTRIBUTION  (Bs: PRIMITIVES ).
  Import Bs.

  Parameter T:Type.
  (** The sequetial view of the distributed data type *)
  Parameter t : T.

  (** The type for indexing element in the sequetial view of the distributed data type *)
  Parameter index : Type.

(** As the distribution may be something heavy to compute, we could need a datatype to store it. *)
  Parameter distribution : Type.
  
  (** Property ensuring that the index denotes a element in the distribution  *)
  Parameter correct_index : index -> distribution -> Prop.
  
(** Using a distribution we can compute the location of an element, if it exists  *)
  Parameter get_proc : forall (d: distribution) (i : index) (i_correct : correct_index i d),  processor. 
 End DISTRIBUTION.


 (**  ** Distribution of types with local index
    a local index is an index in a local part  of a parallel data structure. 
    This type of Distribution enbale to find the local index giving the position of the element in the local structure at processor [p].

*)
Module Type DISTRIBUTION_WITH_LOCAL_INDEX_OPTION (Bs : PRIMITIVES) <:DISTRIBUTION_OPTION Bs.

 Import Bs.
  Include DISTRIBUTION_OPTION Bs.

  (** Index type in a local representation  *)
  Parameter local_index :Type.
  (** Getting the position of an element in the parallel data structure from its posistion in the sequential one   *)
  Parameter get_local_index : distribution -> index ->  option local_index.
  Parameter get_position : distribution -> index -> option (local_index*processor).

  (**  get_position is compatible with get_proc and get_local_index*)
  Parameter position_is_proc_and_local : forall d i, 
    get_position d i =  None /\ get_proc d i= None /\get_local_index d i = None
    \/
    exists li, exists  p, get_position d i =  Some (li,p) /\ get_proc d i = Some p /\ get_local_index d i = Some li.   
End DISTRIBUTION_WITH_LOCAL_INDEX_OPTION.


 (**  ** Distribution of types with local index
    a local index is an index in a local part  of a parallel data structure. 
    This type of Distribution enbale to find the local index giving the position of the element in the local structure at processor [p].

*)
Module Type DISTRIBUTION_WITH_LOCAL_INDEX (Bs : PRIMITIVES) <:DISTRIBUTION Bs.

 Import Bs.
  Include DISTRIBUTION Bs.

  (** Index type in a local representation  *)
  Parameter local_index :Type.

  (** Getting the position of an element in the parallel data structure from its posistion in the sequential one   *)
  Parameter get_local_index : forall  (d:distribution) (i: index) (i_ok:correct_index i d), local_index.
  Parameter get_position : forall   (d:distribution) (i: index) (i_ok:correct_index i d),  (local_index*processor).

  (**  get_position is compatible with get_proc and get_local_index*)
  Parameter position_is_proc_and_local : forall d i i_ok, 
    get_position d i i_ok  = ( get_local_index d i i_ok,get_proc d i i_ok).   

End DISTRIBUTION_WITH_LOCAL_INDEX.


(** ** Distribution of parallel data structure with a notion of local cost
   the cost may be the size of the local par of a parallel datatype, or the time to evaluate a given function on a local element, 
   or even  the time to evaluate the local function, local element can be a function .
   *)
Module Type DISTRIBUTION_WITH_COST  (Bs : PRIMITIVES) <:DISTRIBUTION_NO_INDEX Bs.

 Import Bs.
  Include DISTRIBUTION_NO_INDEX Bs.

  (** Cost of a local element  (most of the times it is its size) *)
  Parameter cost : Type.

  (** Using a distribution we can compute the  cost  of  the elements at a given processor  *)
  Parameter get_cost : distribution -> processor -> cost. 

  (** Using a distribution we can compute the balance of the parallel representation  *)
  Parameter get_balance : distribution -> cost. 
End DISTRIBUTION_WITH_COST.
