Unset Automatic Introduction.    (* coq 8.3 *)
Require Import Arith.
Require Import Bool.
Require Import List.
Require Import Omega.
Require Import Zdiv.
Require Import Euclid.

Require Import Program.
Require Import ProgramCalculation.Tactics.

Require Import LIFO.List.Main.
Require Import LIFO.Nat.Min.    

Require Export  CertifiedBSML.QuickStart.
Require Export SDPP.CorrectParallelisation.Structure.

Module Type  TYPE (Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
(Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
(Parallel : Parallelisation ).

 (* Module Bsml_prop := Bsml_prop Bs. *)
  (* Module Bsml_base := Bsml_base Bs. *)

  Import Bs.
  Import Bsml_Base.
  Import Bsml_Prop.

  (** *Partition of lists among processors *)

  (** list of pids*)
  Definition pids_list := seq 0 bsp_p.


(** list of pids is not nil, we always have at least 1 processor *)
  Lemma pids_list_notnil : pids_list <> nil.
  Proof.
    unfold pids_list. destruct bsp_pSMaxPid as [x eqBSPP]; rewrite eqBSPP.
    simpl. apply sym_not_equal. apply nil_cons.
  Qed.

(** function mapping and folding  over all processor*)

  Definition map_procs (A: Type) (f: nat -> A) :=
    map f pids_list.

  Definition fold_procs (A: Type) (f: A -> nat -> A) (x0: A) :=
    fold_left f pids_list x0.

  Section list_partition.

  (** ** Partitioning of a list of length [len]  into [npart] part   *)
    Variable (nparts: nat).
    Variable (len: nat).

  (** Index of left-most element of part [i]*)
    Definition cut_border (i: nat): nat :=
      let (x, _) := quotient (S nparts) (gt_Sn_O nparts) (len * i) in x.

  (** left-most element index of part [i] is lesser or equal to left-most element index of part [i'] if [i]<=[i']  *)
    Lemma cut_order (i i': nat):
      i <= i' -> (cut_border i <= cut_border i').
    Proof.
      intros i i' leii'.
      unfold cut_border.
      destruct (quotient (S nparts) (gt_Sn_O nparts) (len*i)) as [q1 e].
      destruct (quotient (S nparts) (gt_Sn_O nparts) (len*i')) as [q2 e0].
      destruct e as [r1 H0].
      destruct e0 as [r2 H1].
      destruct H0 as [H2 H3].
      destruct H1 as [H4 H5].

      assert (H : len * i <= len * i').
      apply mult_le_compat_l.
      assumption.
      rewrite H2 in H.
      rewrite H4 in H.
      clear leii' H2 H4.

      elim (le_lt_dec q2 q1). intro leq2q1.
      rewrite le_plus_minus with (m:=q1) (n:=q2) in H.
      rewrite mult_plus_distr_r in H.
      rewrite <- plus_assoc in H.
      apply plus_le_reg_l in H.
      apply le_lt_trans with (p:=S nparts) in H.
      apply le_lt_trans with (n:=(q1-q2)*(S nparts)) in H.
      rewrite mult_comm in H.

      assert (H0:q1-q2<1 -> q1<=q2).
      intro. omega. apply H0. clear H0.
      elim (le_or_lt 1 (q1-q2)).
      intro H'. cut ((S nparts) * (q1-q2) < (S nparts) * (q1-q2)). intro H''.
      elim (lt_irrefl _ H'').
      apply lt_le_trans with (m:=S nparts). assumption.
      rewrite <- mult_1_r with (n:=S nparts).
      rewrite mult_assoc_reverse.
      rewrite mult_1_l.
      apply mult_le_compat_l. assumption.
      trivial.
      omega. omega.
      exact leq2q1.
      intuition.
    Qed.
    Hint Resolve cut_order.

  (** left-most element index of part [0] is [0] *)
    Lemma cut_border_limit_low : cut_border 0 = 0.
    Proof.
      unfold cut_border.
      destruct (quotient (S nparts) (gt_Sn_O nparts) (len*0)) as [q e].
      destruct e as [r H].
      destruct H as [H H0].
      rewrite mult_0_r in H.
      symmetry in H.
      rewrite mult_comm in H.
      simpl in H.
      rewrite <- plus_assoc in H.
      apply plus_is_O in H.
      destruct H.
      assumption.
    Qed.
    Hint Resolve cut_border_limit_low.

  (** left-most element index of part [ nparts+1] is [len] *)
    Lemma cut_border_limit_high : cut_border (S nparts) = len.
    Proof.
      unfold cut_border.
      destruct (quotient (S nparts) (gt_Sn_O nparts) (len*(S nparts))) as [q e].
      destruct e as [r H].
      destruct H as [H H0].
      apply le_antisym.
      
      apply mult_S_le_reg_l with (n:=nparts).
      rewrite mult_comm in H. rewrite H.
      rewrite mult_comm.
      apply le_plus_l.
      
      apply plus_minus in H.
      rewrite <- mult_minus_distr_r in H.
      rewrite <- mult_1_l with (n:=S nparts) in H0.
      rewrite H in H0.
      unfold gt in H0.
      elim (le_or_lt 1 (len-q)). intro leH.
      apply mult_le_compat_r with (p:=S nparts) in leH.
      apply le_lt_trans with (p:=1*(S nparts)) in leH.
      elim (lt_irrefl _ leH). assumption.
      intro leH.
      omega.
    Qed.
    Hint Resolve cut_border_limit_high.

  (** left-most element index of part[i], [i<= nparts+1] is lesser or equal to [len] *)

    Lemma cut_border_range (i: nat):
      i <= S nparts -> cut_border i <= len.
    Proof.
      intros.
      replace (cut_border i <= len)
        with (cut_border i <= cut_border (S nparts)).
      intuition. auto.
    Qed.
    Hint Resolve cut_border_range.

  End list_partition.

  Hint Resolve cut_order cut_border_limit_low cut_border_limit_high cut_border_range.

(** * Partition of a list among processor for BSML *)
  Section bsml_partition.

    Variable (A: Type).

     (** [i]th partition of list [l]*)
    Definition list_partition (l: list A) (i:nat):=
        let len := length l in
          list_part l (cut_border max_pid len i) (cut_border max_pid len (S i)).

    Lemma list_partition_nil  (i:nat): list_partition [] i = [].
      Proof.
        intros i.
        unfold list_partition. simpl. unfold list_part. rewrite skipn_nil. rewrite firstn_nil. reflexivity.
      Qed.

    Lemma partial_partition_merge (l: list A) (i: nat) :
      i < bsp_p ->
      (list_partition l i)++(list_partition l (S i))
      = list_part l (cut_border max_pid (length l) i) (cut_border max_pid (length l) (S (S i))).
    Proof.
      intros l i le_i_bspp.
      unfold list_partition.
      apply list_part_merge.
      split;[
        split;
          auto
        |
          rewrite <- (cut_border_limit_high max_pid);
            auto
      ].
    apply cut_order.
    rewrite <- max_pid_S_bsp_p.
    auto with arith.
  Qed.

(*  Lemma truc : firstn (cut_border len i) = (list_partition l 0) ++ .. ++ (list_partition l i) *)

  (** Merging distributed partitions of a list*)
    Definition partition_merge (l: nat -> list A) :=
      fold_procs  _ (fun x i => x++(l i)) nil.

    Lemma partial_partition_merge_2 (l: list A) (i: nat) :
      i <= max_pid ->
      forall (fp: nat -> list A), (forall n:nat, n<bsp_p -> fp n = list_partition l n) ->
        (fold_left (fun x j => x ++ fp j) (seq 0 (S i)) nil)
        = list_part l 0 (cut_border max_pid (length l) (S i)).
    Proof.
      intros l i le_i_maxpid fp fp_partition. generalize le_i_maxpid.
      induction i as [|i IHi].
      intros le_i_maxpid0. simpl.
      rewrite fp_partition. unfold list_partition.
      auto. rewrite max_pid_S_bsp_p. auto with arith.
      set (f:=fun x j => x ++ fp j) in *|-*.
      intros le_i_maxpid0.
      replace (seq 0 (S (S i))) with (seq 0 (S i) ++ (S i :: nil)).
      rewrite fold_left_app. unfold fold_left at 1.
      unfold f at 1.
      rewrite IHi. rewrite fp_partition. unfold list_partition. rewrite list_part_merge.
      reflexivity.
      split.
      split.
      rewrite <- cut_border_limit_low with (nparts:=max_pid) (len:=length l).
      apply cut_order. intuition.
      apply cut_order. intuition.
      set (len:=length l) in |-* at 2.
      rewrite <- cut_border_limit_high with (nparts:=max_pid) (len:=len).
      subst len.
      apply cut_order. omega.
      rewrite  max_pid_S_bsp_p. auto with arith.
      auto with arith.
      auto with arith.
      change (S i :: nil) with (seq (S i) 1).
      rewrite <- seqApp0. f_equal. intuition.
    Qed.

    Lemma total_fpartition_merge (l: list A) (f: nat -> list A) :
      (forall n, n < bsp_p -> f n = list_partition l n) ->
      partition_merge f = l.
    Proof.
      intros l f f_partition.
      unfold partition_merge.
      unfold fold_procs.
      unfold pids_list.
      rewrite  max_pid_S_bsp_p.
      rewrite partial_partition_merge_2 with (l:=l).
      rewrite cut_border_limit_high.
      unfold list_part.
      simpl.
      rewrite <- minus_n_O.
      apply firstlength.
      auto.
      assumption.
    Qed.

  (** Splitting o Merging = identity*)
    Lemma total_partition_merge (l: list A) :
      partition_merge (list_partition l) = l.
    Proof.
      intros l.
      apply total_fpartition_merge with (f:=list_partition l).
      reflexivity.
    Qed.

    (* Par List Decomposition Scheme *)
    Program Definition map_getInSig_getProcessors:  
      forall (A :Type)(p l:nat)(pproc :p<bsp_p)(l_ok : p+l <=bsp_p) (V : par A) ,
      map  (get V) (InSig.inSig _ (seq p l) _) = map (get V) (firstn l (skipn  p processors)):= _.
    Next Obligation.
      intros A0 p l pproc l_ok V a H.
      apply inSeqLt in H.
      omega.
    Qed.

    Next Obligation.
      intros.
      f_equal.
      simpl.      
      apply (nthSameLengthEqual _ _ firstProcessor).
      rewrite (inSigLength(fun pid : nat => pid < bsp_p)) .
      autorewrite with length.
      rewrite (inSigLength(fun pid : nat => pid < bsp_p)) .
      autorewrite with length.
      rewrite min_minus.
      omega.
      intros n H.
      autorewrite with length in H.
      rewrite (inSigLength(fun pid : nat => pid < bsp_p)) in H .
      autorewrite with length in H.
      rewrite firstn_nth .
      apply processor_eq.
      
      unfold firstProcessor. 
      rewrite skipn_nth.
      assert (H':n+p <bsp_p) by omega.
      erewrite (nthInSigPI(x:=n+p) (m:=bsp_p) (seq p l) n _  H'  ).
      erewrite (nthInSigPI (m:=bsp_p) (seq 0 bsp_p) (n+p) _   H' ).
      simpl.      
      reflexivity.
      repeat rewrite seq_nth.
      simpl. rewrite AlgebraClasses.commutative. auto. 
      auto.
      rewrite seq_nth;
      omega.
      rewrite  (inSigLength(fun pid : nat => pid < bsp_p)). 
      autorewrite with length.
      omega.
      assumption.
    Qed.

Implicit Arguments map_getInSig_getProcessors [A].



Lemma parallel_list_nil  : 
    forall plst : par (list A),
        listOfParList plst = []
        ->
        forall p,
          get plst p = [].
Proof.
  intros plst H p. 
  unfold Flatten.flatten in H.
   apply (fold_right_app_nil _ H  (get plst p)).
  apply in_map.
  apply inProcessors.
Qed.

(** Expressing listOfParList using [get] and [fold_right]  *)
Lemma listOfParList_fold_left_app_map_get :
  forall  plst : par (list A),
    listOfParList plst = fold_left (fun x y => x ++ y) (map (fun i => get plst i) (proj1_sig processors)) [].
Proof.
  intros plst.
  unfold listOfParList.
  induction (proj1_sig processors).
  reflexivity.
  simpl.
  rewrite fold_left_app_acc.
  rewrite IHl.  reflexivity.
Qed.

(** Expressing listOfParList using [get] and [fold_right]  *)
Lemma listOfParList_fold_right_app_map_get :
  forall  plst : par (list A),
    listOfParList plst = fold_right (fun x y => x ++ y) [] (map (fun i => get plst i) (proj1_sig processors)).
Proof.
  intros plst.
  unfold listOfParList.
  induction (proj1_sig processors).
  reflexivity.
  simpl.
f_equal.  
Qed.

Lemma parallel_list_start_somwhere  : 
    forall plst : par (list A),
      forall (a : A) (lst : list A),
        listOfParList plst = a :: lst 
        ->
        exists p,
          (exists l, get plst p = a::l) 
          /\ forall p', (proj1_sig p') < (proj1_sig p) -> get plst p' = [].
Proof.
  intros plst a lst H.
  rewrite listOfParList_fold_left_app_map_get in H.
  elim (fold_left_app_first_elem _ H).
  intros lst' H0. elim H0; clear H0; intros lnil H0; elim H0; clear H0; intros ll' map_decomp. 
  destruct map_decomp as [map_decomp procs_with_nil]. 
  assert (length  (lnil ++ [a :: lst'] ++ ll') = bsp_p).
  rewrite <- map_decomp.
  autorewrite with length.
  reflexivity.
  assert (length lnil < bsp_p).
  autorewrite with length in H0.
  simpl in H0.
  omega.
  exists (processorOfNat  (length lnil ) H1).
  split.
  exists lst'.
  rewrite <- (nthProcessors (processorOfNat (length lnil) H1) firstProcessor).
  rewrite <- map_nth. 
  change (map (get plst) (proj1_sig processors)) with (map (fun i : processor => get plst i) (proj1_sig processors)).
  rewrite map_decomp.
  simpl.
  rewrite  app_nth2.
  replace (length lnil - length lnil) with 0 by (auto with arith).
  reflexivity.
  auto with arith.
  intros p' H2.
  simpl in H2.
  rewrite <- (nthProcessors p' firstProcessor).
  rewrite <- map_nth. 
  change (map (get plst) (proj1_sig processors)) with (map (fun i : processor => get plst i) (proj1_sig processors)).
  rewrite map_decomp.
  rewrite  (app_nth1 _ _ _  H2).
  apply (procs_with_nil _ (nth_In lnil (get plst firstProcessor) H2)).
Qed.

Lemma parallel_list_stop_somwhere  : 
    forall plst : par (list A),
      forall (a : A) (lst : list A),
        listOfParList plst = lst ++[a]
        ->
        exists p,
          (exists l, get plst p = l++[a]) 
          /\ forall p', (proj1_sig p) < (proj1_sig p') -> get plst p' = [].
Proof.
  intros plst a lst H.
  rewrite listOfParList_fold_left_app_map_get in H.
  elim (fold_left_app_last_elem  _ _ _ H).
  intros lst' H0. elim H0; clear H0; intros lnil H0; elim H0; clear H0; intros ll' map_decomp. 
  destruct map_decomp as [map_decomp procs_with_nil]. 
  assert (length  (ll' ++ [ lst'++[a]] ++ lnil) = bsp_p).
  rewrite <- map_decomp.
  autorewrite with length.
  reflexivity.
    assert (length ll' < bsp_p).
  autorewrite with length in H0.
  simpl in H0.
  omega.

  exists (processorOfNat  (length ll' ) H1).
  split.
  exists lst'.
  rewrite <- (nthProcessors (processorOfNat (length ll') H1) firstProcessor).
  rewrite <- map_nth. 
  change (map (get plst) (proj1_sig processors)) with (map (fun i : processor => get plst i) (proj1_sig processors)).
  rewrite map_decomp.
  simpl.
  rewrite  app_nth2.
  replace (length ll' - length ll') with 0 by (auto with arith).
  reflexivity.
  auto with arith.

  intros p' H2.
  simpl in H2.
  rewrite <- (nthProcessors p' firstProcessor).
  rewrite <- map_nth. 
  change (map (get plst) (proj1_sig processors)) with (map (fun i : processor => get plst i) (proj1_sig processors)).
  rewrite map_decomp.
  rewrite app_assoc.
  rewrite  (app_nth2 _ _ _ ).
  assert (proj1_sig p' - length (ll' ++ [lst' ++ [a]]) < length lnil) .
  autorewrite with length. 
  autorewrite with length in H0. 
  destruct p'.  
  simpl in *.  
  omega.
  apply (procs_with_nil _ (nth_In lnil (get plst firstProcessor ) H3)).
  
  autorewrite with length. simpl. omega.
Qed.


  Lemma listOfParList_decomp : 
    forall (p: processor) (pl : par (list A)), 
        listOfParList pl =  ( (fold_left  (fun x i => x++(get pl i))  (firstn (proj1_sig  p) (proj1_sig processors)) nil )++get pl p ++(fold_left  (fun x i => x++(get pl i))  (skipn (proj1_sig p +1) (proj1_sig processors)) nil) ). 
    Proof.
      intros p pl.
      (* induction p using (processor_ind). *)
      unfold listOfParList.
      replace (proj1_sig processors) with (  (firstn (`p) (`processors)) ++ [p] ++ (skipn (`p+1) (`processors))).
      repeat rewrite map_app.
      repeat rewrite Flatten.flattenApp.
      f_equal.
      rewrite flattenFoldLeftApp.
      rewrite firstn_app_length.
      rewrite <-  fold_left_compose_right_map.
      reflexivity.
      autorewrite with length.
      destruct p.
      simpl. rewrite min_minus.
      omega.
     
      f_equal.
      simpl.      
      rewrite app_nil_r.
      reflexivity.
      rewrite flattenFoldLeftApp.
      rewrite <-  fold_left_compose_right_map.
      replace (skipn (`p + 1) (firstn (`p) (`processors) ++ [p] ++ skipn (`p + 1) (`processors)))
        with (skipn (`p + 1) (`processors)).
      reflexivity.
      rewrite <- (AlgebraClasses.associative (eqA:=eq) (op:= @app _)).
      rewrite (skipn_app_length).
      reflexivity.
      autorewrite with length.
      destruct p.
      simpl; rewrite min_minus; omega.      
      simpl.

      (* Decoupage de la liste en trois à mettre dans Properties *)

      assert (H2 : forall a : nat, In a (firstn (proj1_sig p) (seq 0 bsp_p)) -> a < bsp_p).

      intros a H;
      destruct p;  simpl in *.  rewrite firstn_seq in H. apply inSeq0Lt in H.
      rewrite min_minus in H; omega. 

      rewrite (firstnInSig  (seq 0 bsp_p)   _ H2 ).
      2 : destruct p; simpl; autorewrite with length; auto.

      assert (H2' : forall a : nat, In a (skipn (proj1_sig p +1) (seq 0 bsp_p)) -> a < bsp_p).
      intros a H;
      destruct p;  simpl in *.  rewrite skipn_seq in H. apply inSeqLt in H.
      omega.

      rewrite (skipnInSig _  (seq 0 bsp_p)   _ H2' ).

      change (p :: inSig (fun x : nat => x < bsp_p) (skipn (`p + 1) (seq 0 bsp_p)) H2')
        with ([p] ++ (inSig (fun x : nat => x < bsp_p) (skipn (`p + 1) (seq 0 bsp_p)) H2')).
      destruct p  as [p p_proc].
      assert (forall x, In x [p] -> x < bsp_p).
      simpl.
      intros x H. omega.
      rewrite <- (AlgebraClasses.associative (eqA:=eq)).
      
      simpl in *.
      assert (  [exist (fun pid : nat => pid < bsp_p) p p_proc] = (inSig _ [p] H)).
      simpl. f_equal. apply processor_eq. reflexivity.
      
      unfold processor. rewrite H0.
      assert (H3 :forall x, In x   (firstn p (seq 0 bsp_p) ++[p]) -> x < bsp_p). 
      intros x H1.
      rewrite in_app_iff in H1. destruct H1.
      apply H2; assumption.
      simpl in H1.
      destruct H1; auto.      

      rewrite (inSig_app  _ _ _ _ H3).
      assert (H3' :forall x, In x   ((firstn p (seq 0 bsp_p) ++[p]) ++  (skipn (p + 1)(seq 0 bsp_p))) -> x < bsp_p) . 
      intros x H1.
      rewrite in_app_iff in H1. destruct H1;auto.
      rewrite (inSig_app  _ _ _ _ H3').
      apply inSig_eq.
      rewrite firstn_seq.        
      rewrite skipn_seq.
      replace (MinMax.min bsp_p p) with p.
      change [p] with [0+p].
      rewrite <- seq_S_app. 
      replace (p+1) with (S p).
      rewrite <- seqAppN.
      f_equal.        
      omega.
      omega.
      rewrite min_minus; omega.
   Qed.



    Ltac simpl_decomposed_lists:=
      match goal with
        [ all_proc_decomp : (proj1_sig processors) = ?lg++?p::?ld
          ,
           decomp_properties : forall p' : processor,
               (proj1_sig p' < proj1_sig ?p ->
                get ?plst p' = [] /\ get ?plst' p' = []) /\
               (proj1_sig p' = proj1_sig ?p ->
                get ?plst p' = ?a :: ?sublst /\ get ?plst' p' = ?sublst) /\
               (proj1_sig ?p < proj1_sig p' ->
                get ?plst p' = get ?plst' p')
          |- _]
        =>
         (rewrite  (fold_left_all_nil (map (fun x : processor => get plst x) lg)) ;[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p' < proj1_sig p) by
                      apply (splitedProcessorsLeftInf lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ get_plst _];
                      rewrite <- (get_plst ); reflexivity]
        )
        ;
        repeat  (rewrite  (fold_left_all_nil (map (fun x : processor => get plst' x) lg));[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p' < proj1_sig p) by
                      apply (splitedProcessorsLeftInf lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ _ get_plst];
                      rewrite <- (get_plst ); reflexivity]
        )
    end.
     

Lemma sub_par_list :
  forall plst : par (list A),
    forall a lst,
      listOfParList plst = a::lst ->
      exists plst', exists p:processor,
        exists sublst,
          listOfParList plst' = lst
          /\
          forall p':processor,
            (proj1_sig p'< proj1_sig p-> (get plst p' = nil /\get plst' p'= nil))
            /\(proj1_sig p'=proj1_sig p -> (get plst p' = a::sublst /\get plst' p' = sublst))
            /\(proj1_sig p< proj1_sig p' ->get plst p' = get plst' p').
Proof.
  intros plst a lst lopl_plst.
  elim (parallel_list_start_somwhere  plst a _ lopl_plst).
      intros p H0. destruct H0 as [get_plst_p p'_nil]. destruct get_plst_p as [ sublst get_plst_p].
      pose (plst':= mkpar ( fun p' => if lt_dec (proj1_sig p') ( proj1_sig p) then  [] else if le_dec (proj1_sig p') (proj1_sig p) then sublst else (get plst p' ))).
      rename plst' into plst'_w_spec.
      pose (plst':= proj1_sig plst'_w_spec).
      exists  plst' p sublst.
      (* we probably need the right part in the left part of the conjunction *)
      assert ( right_part : (forall p' : processor,
        (proj1_sig p' < proj1_sig p ->
          get plst p' = [] /\ get plst' p' = []) /\
        (proj1_sig p' = proj1_sig p ->
          get plst p' = a :: sublst /\ get plst' p' = sublst) /\
        (proj1_sig p < proj1_sig p' -> get plst p' = get plst' p'))).
      intros p'.
      subst plst'.
      autorewrite with bsml.
      destruct  (lt_dec (`p') ( `p) ) as [p'_lt_p | p'_ge_p]; (split;[|split]);intro p'_p_order;try (contradict p'_p_order;omega).
      (* cas p' < p *)
      split;[apply  (p'_nil p' p'_lt_p) | reflexivity].
      (* cas p'=p *)
      destruct (le_dec (proj1_sig p') (proj1_sig p)); try (contradict p'_p_order; omega).
      split;[|reflexivity].
      rewrite (processor_eq _ _  p'_p_order) in *; rewrite get_plst_p; reflexivity.
      (* cas p'>p *)
      destruct (le_dec (proj1_sig p') (proj1_sig p)); try (contradict p'_p_order; omega).
      reflexivity.

    split;[|assumption].
    (* proof that plst' is a parallel representation of lst*)
    revert lopl_plst.
    repeat rewrite (listOfParList_decomp p).
    repeat rewrite (fold_left_f_inv (fun x => get plst x) ).
    repeat rewrite (fold_left_f_inv (fun x => get plst' x) ).
    repeat rewrite (map_firstn).
    repeat rewrite (map_skipn).
    destructProcessors p.
    repeat rewrite map_app.
    assert (length_lg : forall (pl : par (list A)) , length  (map (fun x : processor => get plst' x) lg) = proj1_sig p ).
                 autorewrite with length.
                 generalize (splitedProcessorsLengthLeft  _ _ _ processors_decomp);  intros length_l1;  symmetry in length_l1.
                 rewrite length_l1.
                 reflexivity.
    assert (length_lg_p : forall (pl : par (list A)) , length  (map (fun x : processor => get plst' x) (lg++[p])) = proj1_sig p + 1 ).
                 autorewrite with length. simpl.
                 generalize (splitedProcessorsLengthLeft  _ _ _ processors_decomp);  intros length_l1;  symmetry in length_l1.
                 rewrite length_l1.
                 reflexivity.
    repeat rewrite <-  (length_lg_p) ;[|try exact plst'] .
    repeat rewrite <-  (length_lg);[|try exact plst'] .
    repeat (rewrite firstn_app_length;[|autorewrite with length;reflexivity]).
    repeat rewrite <- map_app.
    replace (lg++p::ld)with ((lg++[p])++ld) by  (rewrite <-  app_assoc; reflexivity).
    repeat rewrite (map_app _  (lg++[p])).
    repeat (rewrite (skipn_app_length );[|autorewrite with length;reflexivity]).
    simpl_decomposed_lists.

   destruct (right_part p) as [ _ get_p_p'_sup]. destruct get_p_p'_sup as [get_p _ ].
   destruct get_p as [ _ get_plst'_p];[reflexivity|].
   rewrite get_plst'_p, get_plst_p.
   simpl.
   intros lopl_plst.
   injection lopl_plst. clear lopl_plst; intro lopl_plst. rewrite <- lopl_plst.
   f_equal.    f_equal.
   apply map_assumption.
   intros p' in_p'_ld.
   destruct (right_part p') as [ _ get_p_p'_sup]. destruct get_p_p'_sup as [ _ get_p'_sup ].
   symmetry;apply (get_p'_sup  (splitedProcessorsRightSup _ _ _ processors_decomp _ in_p'_ld)).
 Qed.


    (* Ltac simpl_decomposed_lists_rev:= *)
    (*   match goal with *)
    (*     [ all_proc_decomp : (proj1_sig processors) = ?lg++?p::?ld *)
    (*       , *)
    (*        decomp_properties : forall p' : processor, *)
    (*            (proj1_sig ?p < proj1_sig p' -> *)
    (*             get ?plst p' = [] /\ get ?plst' p' = []) /\ *)
    (*            (proj1_sig p' = proj1_sig ?p -> *)
    (*             get ?plst p' = ?sublst ++ [?a] /\ get ?plst' p' = ?sublst) /\ *)
    (*            (proj1_sig p' < proj1_sig ?p -> *)
    (*             get ?plst p' = get ?plst' p') *)
    (*       |- _] *)
    (*     => *)
    (*      (rewrite  (fold_left_all_nil (map (fun x : processor => get plst x) ld)) ;[| *)
    (*       let l :=fresh "l" in let H0:=fresh in *)
    (*         intros l  H0; *)
    (*           rewrite in_map_iff in H0; *)
    (*             let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in *)
    (*               destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def]; *)
    (*               rewrite <- ldef; *)
    (*                 assert ( H0: proj1_sig p < proj1_sig p') by *)
    (*                   apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def); *)
    (*                 let H':= fresh in let get_plst :=fresh in *)
    (*                   destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ get_plst _]; *)
    (*                   rewrite <- (get_plst ); reflexivity] *)
    (*     ) *)
    (*     ; *)
    (*     repeat  (rewrite  (fold_left_all_nil (map (fun x : processor => get plst' x) ld));[| *)
    (*       let l :=fresh "l" in let H0:=fresh in *)
    (*         intros l  H0; *)
    (*           rewrite in_map_iff in H0; *)
    (*             let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in *)
    (*               destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def]; *)
    (*               rewrite <- ldef; *)
    (*                 assert ( H0: proj1_sig p < proj1_sig p') by *)
    (*                   apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def); *)
    (*                 let H':= fresh in let get_plst :=fresh in *)
    (*                   destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ _ get_plst]; *)
    (*                   rewrite <- (get_plst ); reflexivity] *)
    (*     ); *)
    (*   replace (map (fun x : processor => get plst' x) lg) with (map (fun x : processor => get plst x) lg) by( *)
    (*     apply map_assumption; *)
    (*       let p':= fresh "p'" in let in_p'_lg := fresh "in_p'_lg" in *)
    (*         intros p' in_p'_lg; *)
    (*           let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in *)
    (*             destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ]; *)
    (*               apply (get_p'_sup  (splitedProcessorsLeftInf _ _ _ all_proc_decomp _ in_p'_lg))) *)
    (*   ; *)
    (*   try *)
    (*     (match goal with *)
    (*        [p' : processor, *)
    (*          in_p'_ld : In p' ld *)
    (*          |- context e [ get plst' p']]=> *)
    (*        let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in *)
    (*          destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ]; *)
    (*            rewrite (get_p'_sup  (splitedProcessorsRightSup _ _ _ all_proc_decomp _ in_p'_ld)) *)
    (*        | [p' : processor, *)
    (*          in_p'_lg : In p' lg *)
    (*          |- context e [ get plst' p']]=> *)
    (*        let get_p'_inf:= fresh "get_p'_inf" in *)
    (*          let get_plst_p':= fresh "get_plst_p'" in *)
    (*            let get_plst'_p':=fresh "get_plst'_p'" in *)
    (*              destruct (decomp_properties p') as [  get_p'_inf]; *)
    (*                destruct (get_p'_inf (splitedProcessorsLeftInf lg p ld all_proc_decomp _ in_p'_lg)) as *)
    (*                  [ get_plst_p' get_plst'_p']; *)
    (*                  repeat rewrite get_plst'_p' ; repeat rewrite get_plst_p' *)
    (*      end) *)
    (*   end. *)
    Ltac simpl_decomposed_lists_rev:=
      match goal with
        [ all_proc_decomp : (proj1_sig processors) = ?lg++?p::?ld
          ,
           decomp_properties : forall p' : processor,
               (proj1_sig ?p < proj1_sig p' ->
                get ?plst p' = [] /\ get ?plst' p' = []) /\
               (proj1_sig p' = proj1_sig ?p ->
                get ?plst p' = ?sublst ++ [?a] /\ get ?plst' p' = ?sublst) /\
               (proj1_sig p' < proj1_sig ?p ->
                get ?plst p' = get ?plst' p')
          |- _]
        =>
         (rewrite  (fold_left_all_nil (map (fun x : processor => get plst x) ld)) ;[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p < proj1_sig p') by
                      apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ get_plst _];
                      rewrite <- (get_plst ); reflexivity]
        )
        ;
        repeat  (rewrite  (fold_left_all_nil (map (fun x : processor => get plst' x) ld));[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p < proj1_sig p') by
                      apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ _ get_plst];
                      rewrite <- (get_plst ); reflexivity]
        );
      replace (map (fun x : processor => get plst' x) lg) with (map (fun x : processor => get plst x) lg) by(
        apply map_assumption;
          let p':= fresh "p'" in let in_p'_lg := fresh "in_p'_lg" in
            intros p' in_p'_lg;
              let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in
                destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ];
                  apply (get_p'_sup  (splitedProcessorsLeftInf _ _ _ all_proc_decomp _ in_p'_lg)))
      ;
      try
        (match goal with
           [p' : processor,
             in_p'_ld : In p' ld
             |- context e [ get plst' p']]=>
           let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in
             destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ];
               rewrite (get_p'_sup  (splitedProcessorsRightSup _ _ _ all_proc_decomp _ in_p'_ld))
           | [p' : processor,
             in_p'_lg : In p' lg
             |- context e [ get plst' p']]=>
           let get_p'_inf:= fresh "get_p'_inf" in
             let get_plst_p':= fresh "get_plst_p'" in
               let get_plst'_p':=fresh "get_plst'_p'" in
                 destruct (decomp_properties p') as [  get_p'_inf];
                   destruct (get_p'_inf (splitedProcessorsLeftInf lg p ld all_proc_decomp _ in_p'_lg)) as
                     [ get_plst_p' get_plst'_p'];
                     repeat rewrite get_plst'_p' ; repeat rewrite get_plst_p'
         end)
      end.

Lemma sub_par_list_rev :
  forall plst : par (list A),
    forall a lst,
      listOfParList plst = lst++[a] ->
      exists plst', exists p:processor,
        exists sublst,
          listOfParList plst' = lst
          /\
          forall p':processor,
            (proj1_sig p< proj1_sig p'-> (get plst p' = nil /\get plst' p'= nil))
            /\(proj1_sig p'=proj1_sig p -> (get plst p' = sublst++[a] /\get plst' p' = sublst))
            /\(proj1_sig p'< proj1_sig p ->get plst p' = get plst' p').
Proof.
  intros plst a lst lopl_plst.
  elim (parallel_list_stop_somwhere  plst a _ lopl_plst).
      intros p H0. destruct H0 as [get_plst_p p'_nil]. destruct get_plst_p as [ sublst get_plst_p].
      pose (plst':= mkpar ( fun p' => if lt_dec (proj1_sig p) ( proj1_sig p') then  [] else if le_dec (proj1_sig p) (proj1_sig p') then sublst else (get plst p' ))).
      rename plst' into plst'_w_spec.
      pose (plst':= proj1_sig plst'_w_spec).
      exists plst' p sublst.
      (* we probably need the right part in the left part of the conjunction *)
      assert ( right_part : (forall p' : processor,
        (proj1_sig p < proj1_sig p' ->
          get plst p' = [] /\ get plst' p' = []) /\
        (proj1_sig p' = proj1_sig p ->
          get plst p' = sublst++[a] /\ get plst' p' = sublst) /\
        (proj1_sig p' < proj1_sig p -> get plst p' = get plst' p'))).
      intros p'.
      subst plst'.
      autorewrite with bsml.
      
      destruct  (lt_dec (proj1_sig p) ( proj1_sig p') ) as [p'_lt_p | p'_ge_p];(split;[|split]);intro p'_p_order;try (contradict p'_p_order;omega).
      (* cas p' < p *)
      split;[apply  (p'_nil p' p'_lt_p) | reflexivity].
      (* cas p'=p *)
      destruct (le_dec (proj1_sig p) (proj1_sig p')); try (contradict p'_p_order; omega).
      split;[|reflexivity].
      rewrite (processor_eq _ _  p'_p_order) in *; rewrite get_plst_p; reflexivity.
      (* cas p'>p *)
      destruct (le_dec (proj1_sig p) (proj1_sig p')); try (contradict p'_p_order; omega).
      reflexivity.

    split;[|assumption].
    (* proof that plst' is a parallel representation of lst*)
    revert lopl_plst.
    repeat rewrite (listOfParList_decomp p).
    repeat rewrite (fold_left_f_inv (fun x => get plst x) ).
    repeat rewrite (fold_left_f_inv (fun x => get plst' x) ).
    repeat rewrite (map_firstn).
    repeat rewrite (map_skipn).
    destructProcessors p.
    repeat rewrite map_app.
    assert (length_lg : forall (pl : par (list A)) , length  (map (fun x : processor => get plst' x) lg) = proj1_sig p ).
                 autorewrite with length.
                 generalize (splitedProcessorsLengthLeft  _ _ _ processors_decomp);  intros length_l1;  symmetry in length_l1.
                 rewrite length_l1.
                 reflexivity.
    assert (length_lg_p : forall (pl : par (list A)) , length  (map (fun x : processor => get plst' x) (lg++[p])) = proj1_sig p + 1 ).
                 autorewrite with length. simpl.
                 generalize (splitedProcessorsLengthLeft  _ _ _ processors_decomp);  intros length_l1;  symmetry in length_l1.
                 rewrite length_l1.
                 reflexivity.
    repeat rewrite <-  (length_lg_p) ;[|try exact plst'] .
    repeat rewrite <-  (length_lg);[|try exact plst'] .
    repeat (rewrite firstn_app_length;[|autorewrite with length;reflexivity]).
    repeat rewrite <- map_app.
    replace (lg++p::ld)with ((lg++[p])++ld) by  (rewrite <-  app_assoc; reflexivity).
    repeat rewrite (map_app _  (lg++[p])).
    repeat (rewrite (skipn_app_length );[|autorewrite with length;reflexivity]).
     simpl_decomposed_lists_rev.

     destruct (right_part p) as [ _ get_p_p'_sup]. destruct get_p_p'_sup as [get_p _ ].
   destruct get_p as [ _ get_plst'_p];[reflexivity|].
   rewrite get_plst'_p, get_plst_p.
   simpl.
   intros lopl_plst.
   simpl.
   rewrite app_nil_r in lopl_plst.

   repeat rewrite app_assoc in lopl_plst.
   elim (app_inj_tail _ _ _ _ lopl_plst).
   intros H H0.
   rewrite app_nil_r.
   auto.
 Qed.

End bsml_partition.
  Implicit Arguments list_partition [A].
  Implicit Arguments partition_merge [A].
  Implicit Arguments partial_partition_merge_2 [A].
  Implicit Arguments total_partition_merge [A].
  Implicit Arguments listOfParList [A].
  Implicit Arguments  listOfParList_decomp [A].
  Implicit Arguments  sub_par_list [A].
  Implicit Arguments  sub_par_list_rev [A].


Global  Ltac simpl_decomposed_lists:=
    match goal with
      [ all_proc_decomp : (proj1_sig processors) = ?lg++?p::?ld
        ,
        decomp_properties : forall p' : processor,
          (proj1_sig p' < proj1_sig ?p ->
            get ?plst p' = [] /\ get ?plst' p' = []) /\
          (proj1_sig p' = proj1_sig ?p ->
            get ?plst p' = ?a :: ?sublst /\ get ?plst' p' = ?sublst) /\
          (proj1_sig ?p < proj1_sig p' ->
            get ?plst p' = get ?plst' p')
          |- _]
      =>
      try (
        first
          [match goal with
             [
               get_plst_p : get plst p = a :: sublst,
               get_plst'_p : get plst' p = sublst
               |-_] => repeat rewrite get_plst_p ; repeat rewrite get_plst'_p
           end
            |
              let get_plst_p:=fresh "get_plst_p" in let get_plst'_p:=fresh "get_plst'_p" in  let  H':= fresh "H'" in
                destruct (decomp_properties p) as [  _ H'] ;
                  destruct H' as [H' _];
                    destruct H' as [get_plst_p get_plst'_p];[reflexivity|];
                      repeat rewrite get_plst_p; repeat rewrite  get_plst'_p
          ])
      ;
      repeat (rewrite  (fold_left_all_nil (map (fun x : processor => get plst x) lg)) ;[|
        let l :=fresh "l" in let H0:=fresh in
          intros l  H0;
            rewrite in_map_iff in H0;
              let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p' < proj1_sig p) by
                      apply (splitedProcessorsLeftInf lg p ld all_proc_decomp _ p'def);
                        let H':= fresh in let get_plst :=fresh in
                          destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ get_plst _];
                            rewrite <- (get_plst ); reflexivity]
      )
      ;
      repeat  (rewrite  (fold_left_all_nil (map (fun x : processor => get plst' x) lg));[|
        let l :=fresh "l" in let H0:=fresh in
          intros l  H0;
            rewrite in_map_iff in H0;
              let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p' < proj1_sig p) by
                      apply (splitedProcessorsLeftInf lg p ld all_proc_decomp _ p'def);
                        let H':= fresh in let get_plst :=fresh in
                          destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ _ get_plst];
                            rewrite <- (get_plst ); reflexivity]
      );
      replace (map (fun x : processor => get plst' x) ld) with (map (fun x : processor => get plst x) ld) by(
        apply map_assumption;
          let p':= fresh "p'" in let in_p'_ld := fresh "in_p'_ld" in
            intros p' in_p'_ld;
              let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in
                destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ];
                  apply (get_p'_sup  (splitedProcessorsRightSup _ _ _ all_proc_decomp _ in_p'_ld)))
      ;
      try
        (match goal with
           [p' : processor,
             in_p'_ld : In p' ld
             |- context e [ get plst' p']]=>
           let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in
             destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ];
               rewrite (get_p'_sup  (splitedProcessorsRightSup _ _ _ all_proc_decomp _ in_p'_ld))
           | [p' : processor,
             in_p'_lg : In p' lg
             |- context e [ get plst' p']]=>
           let get_p'_inf:= fresh "get_p'_inf" in
             let get_plst_p':= fresh "get_plst_p'" in
               let get_plst'_p':=fresh "get_plst'_p'" in
                 destruct (decomp_properties p') as [  get_p'_inf];
                   destruct (get_p'_inf (splitedProcessorsLeftInf lg p ld all_proc_decomp _ in_p'_lg)) as
                     [ get_plst_p' get_plst'_p'];
                     repeat rewrite get_plst'_p' ; repeat rewrite get_plst_p'
           | [p' : processor,
             in_p'_lg : In p' lg
             |- context e [ get plst p']]=>
           let get_p'_inf:= fresh "get_p'_inf" in
             let get_plst_p':= fresh "get_plst_p'" in
               let get_plst'_p':=fresh "get_plst'_p'" in
                 destruct (decomp_properties p') as [  get_p'_inf];
                   destruct (get_p'_inf (splitedProcessorsLeftInf lg p ld all_proc_decomp _ in_p'_lg)) as
                     [ get_plst_p' get_plst'_p'];
                     repeat rewrite get_plst'_p' ;repeat rewrite get_plst_p'
         end)
    end.



    Global Ltac simpl_decomposed_lists_rev:=
      match goal with
        [ all_proc_decomp : (proj1_sig processors) = ?lg++?p::?ld
          ,
           decomp_properties : forall p' : processor,
               (proj1_sig ?p < proj1_sig p' ->
                get ?plst p' = [] /\ get ?plst' p' = []) /\
               (proj1_sig p' = proj1_sig ?p ->
                get ?plst p' = ?sublst ++ [?a] /\ get ?plst' p' = ?sublst) /\
               (proj1_sig p' < proj1_sig ?p ->
                get ?plst p' = get ?plst' p')
          |- _]
        =>
         try (rewrite  (fold_left_all_nil (map (fun x : processor => get plst x) ld)) ;[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p < proj1_sig p') by
                      apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ get_plst _];
                      rewrite <- (get_plst ); reflexivity]
        )
        ;
        repeat  (rewrite  (fold_left_all_nil (map (fun x : processor => get plst' x) ld));[|
          let l :=fresh "l" in let H0:=fresh in
            intros l  H0;
              rewrite in_map_iff in H0;
                let p':= fresh"p'"in let  lpdef:=fresh  "lpdef" in let ldef:=fresh "ldef" in let p'def:= fresh "p'def" in
                  destruct H0 as [p' lpdef]; destruct lpdef as [ldef p'def];
                  rewrite <- ldef;
                    assert ( H0: proj1_sig p < proj1_sig p') by
                      apply (splitedProcessorsRightSup lg p ld all_proc_decomp _ p'def);
                    let H':= fresh in let get_plst :=fresh in
                      destruct (decomp_properties p') as [ H' _] ; destruct (H' H0) as [ _ get_plst];
                      rewrite <- (get_plst ); reflexivity]
        );
      try replace (map (fun x : processor => get plst' x) lg) with (map (fun x : processor => get plst x) lg) by(
        apply map_assumption;
          let p':= fresh "p'" in let in_p'_lg := fresh "in_p'_lg" in
            intros p' in_p'_lg;
              let get_p_p'_sup:= fresh "get_p_p'_sup" in let get_p'_sup:= fresh "get_p'_sup" in
                destruct (decomp_properties p') as [ _ get_p_p'_sup];  destruct get_p_p'_sup as [ _ get_p'_sup ];
                  apply (get_p'_sup  (splitedProcessorsLeftInf _ _ _ all_proc_decomp _ in_p'_lg)))
      ;
      try
        (match goal with
           [p' : processor,
             in_p'_ld : In p' ld
             |- context e [ get _ p']]=>
           let get_p'_sup:= fresh "get_p'_sup" in
               let get_plst_p':= fresh "get_plst_p'" in
                 let get_plst'_p':=fresh "get_plst'_p'" in
               destruct (decomp_properties p') as [  get_p'_sup _];
                 destruct (get_p'_sup  (splitedProcessorsRightSup _ _ _ all_proc_decomp _ in_p'_ld)) as [get_plst_p' get_plst'_p'];
                   repeat rewrite get_plst_p';
                   repeat rewrite get_plst'_p'
           | [p' : processor,
             in_p'_lg : In p' lg
             |- context e [ get _ p']]=>
           let get_p'_inf:= fresh "get_p'_inf" in
             let get_plst_p':= fresh "get_plst_p'" in
               let get_plst'_p':=fresh "get_plst'_p'" in
                 destruct (decomp_properties p') as [  _ get_p'_inf];
                   destruct (get_p'_inf) as [  _ get_p'_inf];
                   destruct (get_p'_inf (splitedProcessorsLeftInf lg p ld all_proc_decomp _ in_p'_lg)) as
                     [ get_plst_p' get_plst'_p'];
                     repeat rewrite get_plst'_p' ; repeat rewrite get_plst_p'
           | [|- context e [ get _ p]]=>
             let get_p:= fresh "get_p" in
               let get_plst_p:= fresh "get_plst_p" in
                 let get_plst'_p:=fresh "get_plst'_p" in
                   destruct (decomp_properties p) as [  _ get_p];
                     destruct (get_p) as [ get_p _];
                       let p_eq_p:=fresh "p_eq_p" in
                         assert (p_eq_p : (proj1_sig p = proj1_sig p)) by reflexivity;
                           destruct (get_p  p_eq_p) as
                             [ get_plst_p get_plst'_p];
                             repeat rewrite get_plst'_p ; repeat rewrite get_plst_p
         end)
      end.


Global Ltac decomp_one_lopl p plst :=
    repeat rewrite (listOfParList_decomp p plst);
      first [
        (match goal with
           [H : (proj1_sig processors) = ?lg++p::?ld|- _ ]=> try rewrite H in *
         end
        )
        |
          (  destructProcessors p)
      ];
      repeat rewrite (fold_left_f_inv (fun x => get plst x) );
        repeat rewrite (map_firstn);
          repeat rewrite (map_skipn);
            match goal with
              [ H : (proj1_sig processors) = ?lg++p::?ld |- _]
              =>
    (* cutting of maps *)
              replace (skipn (`p+1) (map (fun x : processor => get plst x) (lg ++ p :: ld)))with
                (skipn (`p+1) (map (fun x : processor =>  get plst x) ((lg++[p])++ld))) by  (rewrite <-  app_assoc; reflexivity);
                repeat rewrite (map_app _  (lg++[p]));
                  repeat rewrite (map_app _ _ (p::ld) );
    (* changing firstn and skipn first parameter for matching firstn_app_length and  skipn_app_length *)
                    first [
                      match goal with
                        [ length_lg_p : length  (map (fun x : processor => get plst x) (lg++[p])) = proj1_sig p + 1|-_] => idtac
                      end
                      |
                        let length_lg_p := fresh "length_lg_p"    in
                          assert (length_lg_p : length  (map (fun x : processor => get plst x) (lg++[p])) = proj1_sig p + 1 ) by(
                            autorewrite with length; simpl;
                              let length_l1 := fresh "length_l1" in(
                                assert (length_l1 := splitedProcessorsLengthLeft  _ _ _ H);  symmetry in length_l1;
                                    rewrite length_l1;
                                      reflexivity)
                          )
                    ]
                    ;
                    first [
                      match goal with
                        [length_lg :  (length  (map (fun x : processor => get plst x) lg) = `p )|-_] => idtac
                      end
                      |
                        let length_lg := fresh "length_lg" in
                          assert (length_lg :  (length  (map (fun x : processor => get plst x) lg) = `p )) by(
                            autorewrite with length;
                              let length_l1 := fresh "length_l1" in (
                                assert  (length_l1:=splitedProcessorsLengthLeft  _ _ _ H);  symmetry in length_l1;
                                  rewrite length_l1; reflexivity)
                          )
                    ]
                       ;
                       match goal with
                         [length_lg :  (length  (map (fun x : processor => get plst x) lg) = `p ),
                           length_lg_p : length  (map (fun x : processor => get plst x) (lg++[p])) = proj1_sig p + 1 |-_] =>
                    
                        repeat rewrite <- length_lg_p;
                          repeat rewrite <- length_lg;
                            repeat (rewrite firstn_app_length;[|autorewrite with length;reflexivity]);
                              repeat (rewrite (skipn_app_length );[|autorewrite with length;reflexivity]);
                                repeat rewrite  length_lg_p;
                                  repeat rewrite  length_lg
                                    
                       end
            end
            ;
            try simpl_decomposed_lists
            ;
            try simpl_decomposed_lists_rev
            .

Global Ltac decomp_lopl p :=
repeat  match goal with
          [ |- context e [listOfParList ?plst] ]=>
          decomp_one_lopl p plst
        end.

Global Ltac generate_sub_par_list plst :=
  match goal with
    [ lopl : listOfParList plst = ?a::?l|-_] => idtac
    | [ lopl :  ?a::?l = listOfParList plst |- _] => symmetry in lopl
  end;
  match goal with
    [ lopl : listOfParList plst = ?a::?l|-_] =>
      let plst' := fresh "plst'" in let H:=fresh "H" in destruct (sub_par_list plst a l  lopl) as [plst' H];
        let p:= fresh "p" in destruct H as [p H];
          let sublst:= fresh "sublst" in let lopl_plst'_par_lists_decomp:= fresh "lopl_plst'_par_lists_decomp" in
            destruct H as [sublst lopl_plst'_par_lists_decomp];
              let  lopl_plst':= fresh "lopl_plst'" in
                let  par_lists_decomp:= fresh "par_lists_decomp" in
                  destruct lopl_plst'_par_lists_decomp as [lopl_plst' par_lists_decomp]
  end.

Global Ltac generate_sub_par_list_rev plst :=
  match goal with
    [ lopl : listOfParList plst = ?l++[?a]|-_] => idtac
    | [ lopl :  ?l++[?a] = listOfParList plst |- _] => symmetry in lopl
  end;
  match goal with
    [ lopl : listOfParList plst = ?l++[?a]|-_] =>
      let plst' := fresh "plst'" in let H:=fresh "H" in destruct (sub_par_list_rev plst a l  lopl) as [plst' H];
        let p:= fresh "p" in destruct H as [p H];
          let sublst:= fresh "sublst" in let lopl_plst'_par_lists_decomp:= fresh "lopl_plst'_par_lists_decomp" in
            destruct H as [sublst lopl_plst'_par_lists_decomp];
              let  lopl_plst':= fresh "lopl_plst'" in
                let  par_lists_decomp:= fresh "par_lists_decomp" in
                  destruct lopl_plst'_par_lists_decomp as [lopl_plst' par_lists_decomp]
  end.

(* Gobal Ltac parlist_induction par_list := *)
(*   match goal with *)
(*     [ |- context e [listOfParList par_list] ]=> *)
(*     let seq_list := fresh seq_list in *)
(*       set (seq_list := listOfParList par_list) ; *)
(*         assert (seq_list = listOfParList par_list) by reflexivity; *)
    
    
  (** Decomposing Homomorphisms applied to parlist join *)
  Lemma listOfParList_decomp_homo :
    forall (A: Type) ,
      forall (p: processor) (pl : par (list A)),
          forall (B: Type)  (h : list A -> B) {h_homo : Homomorphism (B:=B) (eqB:=eq) h},
            h (listOfParList pl) =   fold_left  (Homo_op (h:=h))   (listOfParList (proj1_sig (parfun (map (Homo_f (h:=h))) pl))) (h []).
  Proof.
    intros A p pl B h h_homo.
    rewrite (listOfParList_decomp p (proj1_sig (parfun (map (Homo_f (h:=h))) pl))) .
    rewrite (listOfParList_decomp p).
    repeat rewrite(Homomorphism_list_destruct (h:=h)).
    repeat rewrite fold_homomorphism_2.
    rewrite (fold_homomorphism_3 h_homo).
    rewrite (fold_homomorphism_3 h_homo  (get pl p)).
    rewrite (fold_homomorphism_3 h_homo(fold_left
      (fun (x : list A) (i : processor) => x ++ get pl i)
      (skipn (`p + 1) (proj1_sig processors)) [])).
     repeat rewrite (fold_left_compose_right_map' _ (get (proj1_sig _) )).  
    repeat rewrite (fold_left_compose_right_map' _ (get pl) ).
    repeat rewrite fold_left_map_app.
    repeat rewrite map_map. unfold parfun. unfold replicate.
    assert ( Hyp_map_assum :forall p' : processor,
      In p'  ( firstn (`p) (proj1_sig processors)) ->
      (get (
        proj1_sig
        (apply (proj1_sig (mkpar (fun _ : processor => map (Homo_f (h:=h))))) pl))  p' = (fun x : processor => map (Homo_f (h:=h)) (get pl x) ) p')).
             intros p' H.
             autorewrite with bsml.
             reflexivity. simpl proj1_sig in *.
     rewrite (List.Map.map_assumption  processor (list B)
       _
       (* (get *)
       (*   (`(apply (`(mkpar (fun _ : processor => map Homo_f(h:=H))))   *)
       (*     pl) *)
       (*   ) *)
       (* ) *)
       _
       (fun x : processor => map (Homo_f (h:=h)) (get pl x) )
       Hyp_map_assum
     ).
   assert ( Hyp_map_assum_right :forall p' : processor,
       In p'  (skipn (`p + 1) (proj1_sig processors)) ->
       (get
         (proj1_sig
           (apply (
             proj1_sig
             (mkpar (fun _ : processor => map (Homo_f (h:=h))))) pl)
         )  p'
         =
         (fun x : processor => map (Homo_f (h:=h)) (get pl x) ) p'))
     by(  intros p' H;  autorewrite with bsml; reflexivity  ).
   simpl proj1_sig in *.
   rewrite (List.Map.map_assumption processor (list B)  _ _ _ Hyp_map_assum_right).
   autorewrite with bsml.
  reflexivity.
Qed.

Lemma in_listOfParList  (A:Type)   (x : A) (plst : par (list A)) :
  (exists p, In x (get plst p)) <->( In x (listOfParList plst)).
Proof.
  intros A0 x plst.
  split.
  (* -> *)
  intro exist_p.
  destruct exist_p as [p in_x_plst_p].
  decomp_lopl p.
  apply in_or_app.
  right.
  apply in_or_app.
  left.
  assumption.
  (* <- *)
  intros H.
  rewrite listOfParList_fold_right_app_map_get in H.
  induction (proj1_sig processors).
  simpl in H. elim H.
  simpl in H.
  generalize ( in_app_or _ _  x H).
  intros H0.
  destruct H0.
  exists a. assumption.
  apply IHl.
  assumption.
Qed.
Implicit Arguments in_listOfParList [A].


  
(** The length of the sequential representation is the sum of the local lengths  *)
Lemma listOfParList_length   (A:Type) (plst : par (list A)) : 
  length (listOfParList plst) = fold_left (fun a b => a + b) (map (get (proj1_sig (parfun (@length A ) plst))) (proj1_sig processors)) 0. 
Proof.
  intros A plst.
  rewrite   listOfParList_fold_left_app_map_get.
  autorewrite with length.
  f_equal.
  rewrite map_map.  
  apply map_ext.
  intros a.
  autorewrite with bsml bsmlbase.
  reflexivity.
Qed.

Hint Rewrite listOfParList_length : length.

  Import Parallel.
  Program Definition partition_list (A:Type) (l: list A) :par( list A) := mkpar  ( fun (p :processor) => list_partition l (proj1_sig p)).
  Implicit Arguments partition_list [A].

Global Instance list_partitionnable {A: Type}   : Partitionnable (list A) :={
    parallel_type := par (list A); 
      join := (fun l => @listOfParList A l);
      partition := (fun l => partition_list l)
    }.
  Proof.
    intros a.
        
    rewrite <- (total_partition_merge  a) at 2. 
    unfold partition_merge.
    unfold fold_procs.
    unfold pids_list.
    replace bsp_p with (S (proj1_sig lastProcessor)).
    listOfParListInduction.

    unfold partition_list.
    simpl.   
    autorewrite with bsml.
    reflexivity.
    rewrite  fold_left_compose_right_map'.
    replace  (seq 0 (S (`S_p))) with  ((seq 0 (proj1_sig S_p))++[(proj1_sig S_p)] ).
    Focus 2. replace (S (`S_p) )with ( (`S_p) + 1). rewrite  seqApp0 . reflexivity.
    simpl ; omega.
    rewrite map_app.
    rewrite fold_left_app.
    repeat rewrite  <- fold_left_compose_right_map'.
    rewrite IHp'.
    simpl fold_left.
  f_equal.
unfold partition_list; autorewrite with bsml.
reflexivity.

simpl. generalize bsp_pNotZero. intros H. omega.
  Defined.
End TYPE.

Module MakePartition (Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
(Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
(Parallel : Parallelisation ).
  Include  TYPE Bs Bsml_Prop ParSig Bsml_Base Parallel.
End MakePartition.