(** * Extension of a totally odered set with boundaries
  *)
Unset Automatic Introduction.    (* coq 8.3 *)
Require Import SDPP.Misc.total_order.

Module Type BoundedSig.

Parameter (A : Type).
Parameter R : relation A.
Parameter preoR : PreOrder R.
Parameter totR : TotalOrder R (preo:=preoR).

Transparent A R preoR totR.
End BoundedSig.


Module BoundedSets (M:BoundedSig).
Export M.

(** Elements of a bounded set are either Min, Max, or an element of A.  *)
Inductive BoundedSet  : Type :=
  | Min : BoundedSet
  | Max : BoundedSet
  | Elmt : A-> BoundedSet.

(** Extending of a total order over A the bounded set
   Min is smaller than any element, and all other elements are greater than Min
   Max is greater than any element, and all other elements are smaller than Min
   Order over elements of A is kept.
.  *)
Program Definition makeTotalOrderBoundedSet (R:relation A) {preo : PreOrder R} {tot : TotalOrder R}  : relation BoundedSet := 
  fun x y : BoundedSet => 
      match (x,y) with
      | (Min , _) =>  True
      | (_ , Max) => True
      | (Max , _) => False
      | (_ , Min) => False
      | (Elmt x' , Elmt y') =>  R x' y'
    end.
Next Obligation.
  split;intro w. assert (H: Max <> Elmt Anonymous0) by discriminate. discriminate.
  assert (H: Min <> Elmt Anonymous0) by discriminate. discriminate.
Qed.

Definition BSOrder :=makeTotalOrderBoundedSet R (preo:=preoR) (tot:=totR).

Instance PreOrderR :PreOrder R.
exact preoR.
Qed.

(**  R is a TotalOrder *)
Program Instance TotalOrderR :TotalOrder R.
Next Obligation.
unfold TotalRelation ; intros x y.
apply ((TotalOrder_totality (R:=R) _  _ (preo:=preoR )(TotalOrder := totR))).
Qed.
Next Obligation.
unfold PartialOrder.
unfold relation_equivalence ; unfold relation_conjunction.
unfold predicate_equivalence. unfold predicate_intersection. simpl.
unfold Basics.flip.
intros x x0.
apply ((TotalOrder_PartialOrder (R:=R) x  x0 (preo:=preoR )(TotalOrder := totR))).
Qed.

(**  BSOrder is a PreOrder *)
Program Instance PreOrderBoundedSet_OK : PreOrder (BSOrder).
Next Obligation.
  unfold BSOrder. unfold makeTotalOrderBoundedSet.
  unfold Reflexive.
  intro x ; case x; trivial.
  intro a. 
  reflexivity.
Qed.
Next Obligation.
  unfold BSOrder. unfold makeTotalOrderBoundedSet.
  unfold Transitive.
  intros x y z.
  case x; case y; case z; trivial.
  intros a a0 H H0. elim H.
  intros a a0 H H0. elim H0.  
  intros a a0 a1 H H0. rewrite H. rewrite H0. reflexivity.
Qed.

(**  BSOrder is a TotalOrder *)
Program Instance TotalOrderBoundedSet_OK : TotalOrder (BSOrder).
Next Obligation.
  unfold TotalRelation. unfold BSOrder. unfold makeTotalOrderBoundedSet.
  intros x y. case x; case y; auto.
  intros a a0.
  apply totality.  
Qed.
Next Obligation.
unfold BSOrder. unfold makeTotalOrderBoundedSet.
unfold PartialOrder. unfold relation_equivalence ; unfold relation_conjunction.
unfold predicate_equivalence. unfold predicate_intersection. simpl.
intros x x0. case x; case x0; split;
 ((split; unfold Basics.flip; auto)||(intro H;elim H)|| auto).
discriminate.
repeat (intros H0 H1;
unfold Basics.flip in *; auto; elim H1
).
discriminate.
intros H0 H1;
unfold Basics.flip in *; auto ; elim H1. 
discriminate.
intros H0 H1;
unfold Basics.flip in *; auto; elim H0. 
discriminate.
intros H0 H1;
unfold Basics.flip in *; auto; elim H0. 
discriminate.
intros H0 H1;
unfold Basics.flip in *; auto; elim H0. 
discriminate.
intros H0 H1;
unfold Basics.flip in *; auto; elim H1.
injection H.
intro rew; rewrite rew;
reflexivity.
injection H.
intro rew; rewrite rew.
reflexivity.
intros H0 H1;
unfold Basics.flip in *; auto.
assert (a = a0).
generalize (TotalOrder_PartialOrder (R:=R)). 
unfold PartialOrder. unfold relation_conjunction ; unfold relation_equivalence; unfold Basics.flip.
unfold predicate_intersection ; unfold predicate_equivalence. simpl.
intro H2. generalize (H2 a0 a).
intro H3; destruct H3. symmetry.
apply (H4 H).
rewrite H2.
reflexivity.
Qed.


(**  BSOrder is decidable if R is decidable *)
Definition BSOrder_dec `{R_dec : forall a a0 : A, {R a a0}+{~ R a a0}} : 
  forall a1 a2 :BoundedSet, {BSOrder  a1 a2}+{~(BSOrder  a1 a2)}.
intros R_dec a1 a2.
case a1; case a2 ; auto.
left. reflexivity.
left. unfold BSOrder. unfold makeTotalOrderBoundedSet. trivial.
intros a. left;firstorder.
left ; reflexivity.
intros a. left;firstorder.
intros a a0. 
unfold BSOrder. unfold makeTotalOrderBoundedSet.
exact (R_dec a0 a).
Qed.

End BoundedSets.
