 Unset Automatic Introduction.    (* coq 8.3 *)
Require Import List.
(* Require Import Arith. *)

Require Export Skel.BH.bh.
Require Export Skel.BH.bh_bsml.

Require Import LIFO.List.Main.
Require Import LIFO.Algebra.AlgebraClasses.
Require Import LIFO.List.Homomorphism.

Module Type TYPE(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) .

  Export Bsml_bh.
  Export Parallel.

  (** *BH_PROP Class  
     this Type Classes defined the  properties needed for a function to be BH.
     *)
  Class BH_PROP {L A R B : Type} (bh : list A -> L -> R -> list B):=
  {
    BH_PROP_k :  L -> A -> R ->  B;
    BH_PROP_gl :  list A -> L ;
    BH_PROP_opl :  L -> L -> L;
    BH_PROP_gr :  list A -> R ;
    BH_PROP_opr :  R -> R -> R;
    BH_PROP_nil :  forall l r,  bh [] l r = [] ;
    BH_PROP_singleton :  forall a l r,  bh [a] l r = [ BH_PROP_k l a  r ] ;
    BH_PROP_append :  forall (x y : list A)  l r,  
              bh (x++y) l r = 
              bh x l (BH_PROP_opr (BH_PROP_gr y) r) ++ bh y (BH_PROP_opl l (BH_PROP_gl x) ) r;
    BH_PROP_left_homomorphism : Homomorphism' BH_PROP_opl BH_PROP_gl ;
    BH_PROP_right_homomorphism : Homomorphism' BH_PROP_opr BH_PROP_gr 
  }.

  (** *BH Class  is a specialisation of BH_General where left and right (summaries) parameters   are identities for opl and opr operators
     *)
  Class BH_Special {A B L R : Type}  (f : list A -> list B) (l:L) (r:R) :=
    { BHS_k : L -> A -> R -> B;
      BHS_gl : list A -> L;
      BHS_opl : L -> L -> L;
      BHS_gr: list A -> R;
      BHS_opr: R -> R -> R;
      BHS_spec_match : forall lst : list A, 
        f lst = bh_comp  BHS_k BHS_gl BHS_opl BHS_gr BHS_opr l  lst r ;
      BHS_parallelisation : par(list A) ->par( list B):= 
        (fun plst => bh_bsml_opt_comp  _ _ _ _ BHS_k BHS_gl BHS_opl BHS_gr BHS_opr l plst r);
      BHS_left_homomorphism : Homomorphism' BHS_opl BHS_gl ;
      BHS_right_homomorphism : Homomorphism' BHS_opr BHS_gr 
    }.


  (** *BH Class  
     this Type Classes defined the general construction of a parallel programm using BH.
     It includes :
     - properties over operators
     - a proof that the bh computation using those operators match the spec f
     - a term which is the parallel program   
     *)
  Class BH_General {L A R B : Type} (f : L -> R -> list A -> list B) :=
  {
    BHG_k : L -> A -> R -> B;
    BHG_gl : list A -> L;
    BHG_opl : L -> L -> L;
    BHG_gr: list A -> R;
    BHG_opr: R -> R -> R;
    BHG_spec_match : forall l r lst , f l r lst =  bh_comp  BHG_k BHG_gl BHG_opl BHG_gr BHG_opr l lst r ;
    BHG_parallelisation := (fun  l r lst  =>   bh_bsml_opt_comp  L A R B BHG_k BHG_gl BHG_opl BHG_gr BHG_opr l lst r) : L -> R -> par(list A) ->par( list B) ; 
    BHG_left_homomorphism : Homomorphism' BHG_opl BHG_gl ;
    BHG_right_homomorphism : Homomorphism' BHG_opr BHG_gr 
  }.


(** *[BH_General] give a correct parallel implementation of any function [f] which is [BH_General]   *)
  Global  Program Instance BH_General_Parallel  `(bh : BH_General L A R B  f) l r  : Parallel (f l r ) (R:=eq) (BHG_parallelisation (BH_General:=bh) l r ). 

  Next Obligation.
    destruct bh.
    unfold BHG_parallelisation.
    intros l r ap.
    rewrite (BHG_spec_match0 l r ).
    apply (bh_bsml_opt_comp_bh); firstorder.
  Defined.

(** *A fonction declared [BH_General] is thus easily parallelizable   *)
  Global  Program Instance BH_General_Parallelizable  `(bh : BH_General L A R B  f) l r  : Parallelizable (f l r ) (R:=eq) :={
    parallel:=(BHG_parallelisation (BH_General:=bh) l r );
    spec_match:=parallel_spec_match}.


(** *[BH_Special] give a correct parallel implementation of any function [f] which is [BH_Special]   *)
  Global Program Instance BHS_Parallel  `(bh : BH_Special A B L R f l r) : Parallel f (R:=eq) (BHS_parallelisation (BH_Special:=bh) ).
Next Obligation.
    destruct bh.
    unfold BHS_parallelisation.
    intros ap.
    rewrite (BHS_spec_match0 (join ap)). 
    simpl. 
    rewrite bh_bsml_opt_comp_bh; simpl; firstorder.
  Defined.

(** *For a function which is a BSP Homomorphism ([BH_PROP]) it is always possible to build an instance of [BH_General]   *)
Global Instance BHG_BH_Prop `(f :list A -> L -> R -> list B)  `{bh_prop : BH_PROP L A R B f}
                                                      `{Homomorphism' A  R  eq (BH_PROP_opr) (BH_PROP_gr)} 
                                                      `{Homomorphism' A  L  eq (BH_PROP_opl(BH_PROP:=bh_prop)) (BH_PROP_gl(BH_PROP:=bh_prop))} : 
                                                      BH_General 
                                                       (fun l r la => f la l r):={
                                                         BHG_k:= BH_PROP_k (BH_PROP:=bh_prop);
                                                      BHG_gl:= BH_PROP_gl (BH_PROP:=bh_prop);
                                                      BHG_opl:=BH_PROP_opl (BH_PROP:=bh_prop);
                                                      BHG_gr:=BH_PROP_gr (BH_PROP:=bh_prop);
                                                      BHG_opr:=BH_PROP_opr (BH_PROP:=bh_prop)}.
Proof.
  destruct bh_prop. simpl in *.
  intros;  revert l r;  revert lst.
  apply (homomorphism_induction A (fun lst=>  (forall l r,  f lst l r =
    bh_comp BH_PROP_k0 BH_PROP_gl0 BH_PROP_opl0 BH_PROP_gr0 BH_PROP_opr0 l lst
    r) )).
  intros l r.
  simpl.
  generalize (BH_PROP_append0 [] [] l r).
  intro f_app_nil_nil. simpl in f_app_nil_nil.
  rewrite left_neutral in f_app_nil_nil .
  rewrite (right_neutral)  in f_app_nil_nil .
  rewrite (app_l_l'_nil  _ _ _ f_app_nil_nil).  
  reflexivity.

  intros a l r .
  rewrite BH_PROP_singleton0.
  unfold bh_comp. simpl. unfold bh_nth. simpl. rewrite left_neutral. rewrite right_neutral.
  reflexivity.

  intros l l' IHl IHl' l0 r.
  rewrite  BH_PROP_append0.
  rewrite IHl.
  rewrite IHl'.
  rewrite bh_hom.
  reflexivity.
  
  firstorder.
  firstorder.
  firstorder.  
  firstorder.  
Defined.

 (** *[bh_comp] is parallelized by [bh_bsml_opt_comp]  *)
Global Instance bh_compParallel
  {L A R B : Type} {k: L -> A -> R -> B} 
  {gl: list A -> L} {opl: L -> L -> L} {hl:Homomorphism' opl gl}
  {gr: list A -> R} {opr: R -> R -> R} {hr:Homomorphism' opr gr }
  : Parallel (R:=eq)
  (fun lst =>bh_comp k gl opl gr opr l lst r)
  (fun lst  =>   bh_bsml_opt_comp  L A R B k gl opl gr opr l lst r).
Proof.
  intros L A R B k gl opl hl gr opr hr l r.
  constructor.
  intros par_a.   
  assert (homgl : is_homomorphism A L gl opl) by firstorder.
  assert (homgr :is_homomorphism A R gr opr) by firstorder.
  apply (bh_bsml_opt_comp_bh L A R B k gl opl gr opr  homgl homgr).
Defined.


  (** *For a function which is a BSP Homomorphism ([BH_PROP]) it is always possible to build a parallel implemantion 
    using [bh_bsml_opt_comp_bh]
    *)
Global Instance BH_PROP_Parallel
  `(f :list A -> L -> R -> list B)  `{bh_prop : BH_PROP L A R B f}  l r 
  : Parallel (R:=eq)
  (fun lst =>f lst l r)
  (fun lst  =>   bh_bsml_opt_comp  L A R B  BH_PROP_k BH_PROP_gl BH_PROP_opl BH_PROP_gr BH_PROP_opr l lst r).
Proof.
  intros A L R B f bh_prop l r.
  constructor.
  
  destruct bh_prop.
  intros par_a.
   rewrite (parallel_spec_match  (Parallel:=bh_compParallel)).
   symmetry;
     simpl in *.
  revert l r.
  set (lst :=(listOfParList par_a)).
  generalize lst.
  apply (homomorphism_induction A (fun lst=>  (forall l r,  f lst l r =
    bh_comp BH_PROP_k0 BH_PROP_gl0 BH_PROP_opl0 BH_PROP_gr0 BH_PROP_opr0 l lst
    r) )).
  intros l r.
  simpl.
  generalize (BH_PROP_append0 [] [] l r).
  intro f_app_nil_nil. simpl in f_app_nil_nil.
  rewrite left_neutral in f_app_nil_nil .
  rewrite (right_neutral)  in f_app_nil_nil .
  rewrite (app_l_l'_nil  _ _ _ f_app_nil_nil).  
  reflexivity.

  intros a l r .
  rewrite BH_PROP_singleton0.
  unfold bh_comp. simpl. unfold bh_nth. simpl. rewrite left_neutral. rewrite right_neutral.
  reflexivity.

  intros l l' IHl IHl' l0 r.
  rewrite  BH_PROP_append0.
  rewrite IHl.
  rewrite IHl'.
  rewrite bh_hom.
  reflexivity.

  firstorder.
  firstorder.
Defined.


  (** * A BSP Homomorphism ([BH_PROP]) is thus parallelizable    *)
Global Instance BH_PROP_Parallelizable  `(f :list A -> L -> R -> list B)  `{bh_prop : BH_PROP L A R B f}  l r   : Parallelizable (R:=eq)
  (fun lst =>f lst l r).
Proof.
  intros A L R B f bh_prop l r.
  eapply (parallel_parallelizable _  (H:=( BH_PROP_Parallel _ _ _ ))). 
Defined.


  (** * [BH_Special]) function are parallelizable  *)
Global Instance BHS_Parallelizable  `(bh : BH_Special A B L R f l r) : Parallelizable (R:=eq) f. 
Proof.
  intros A B L R f l r bh.
apply (parallel_parallelizable (R:=eq) f  (fp:=(BHS_parallelisation (f:=f)))) .
Defined. 



Hint Unfold join : Parallel.
Hint Unfold list_partitionnable : Parallel.
Hint Unfold BHG_parallelisation : Parallel.
Hint Unfold BHS_parallelisation : Parallel.
Hint Unfold BH_General_Parallelizable : Parallel.
Hint Unfold BHS_Parallelizable : Parallel.
Hint Unfold BH_PROP_Parallelizable : Parallel.
Hint Unfold BHG_spec_match : Parallel.
Hint Unfold BHS_spec_match : Parallel.
Hint Unfold bh_compParallel : Parallel.
Hint Unfold bh_bsml_opt_comp : Bh.  
End TYPE.

Module Make(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) .
  
  Include TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh.

End Make.

